"""State for the SW1L solver (:mod:`fluidsim.solvers.sw1l.state`)
=================================================================

.. autoclass:: StateSW1L
   :members:
   :private-members:

"""

import numpy as np

from fluidsim.base.setofvariables import SetOfVariables
from fluidsim.base.state import StatePseudoSpectral

from fluiddyn.util import mpi


class StateSW1L(StatePseudoSpectral):
    """State for the solver sw1l.

    Contains the variables corresponding to the state and handles the
    access to other fields for the solver SW1L.

    .. inheritance-diagram:: StateSW1L

    """

    @staticmethod
    def _complete_info_solver(info_solver):
        """Complete the ParamContainer info_solver.

        This is a static method!
        """
        info_solver.classes.State._set_attribs(
            {
                "keys_state_spect": ["ux_fft", "uy_fft", "eta_fft"],
                "keys_state_phys": ["ux", "uy", "eta", "rot"],
                "keys_computable": [],
                "keys_phys_needed": ["ux", "uy", "eta"],
                "keys_linear_eigenmodes": ["q_fft", "a_fft", "d_fft"],
            }
        )

    def compute(self, key, SAVE_IN_DICT=True, RAISE_ERROR=True):
        """Compute and return a variable."""
        it = self.sim.time_stepping.it
        if key in self.vars_computed and it == self.it_computed[key]:
            return self.vars_computed[key]

        if key == "Jx":
            ux = self.state_phys.get_var("ux")
            eta = self.state_phys.get_var("eta")
            h = 1 + eta
            result = h * ux
        elif key == "Jy":
            uy = self.state_phys.get_var("uy")
            eta = self.state_phys.get_var("eta")
            h = 1 + eta
            result = h * uy
        elif key == "Jx_fft":
            Jx = self.compute("Jx")
            result = self.oper.fft2(Jx)
        elif key == "Jy_fft":
            Jy = self.compute("Jy")
            result = self.oper.fft2(Jy)
        elif key == "rot_fft":
            ux_fft = self.state_spect.get_var("ux_fft")
            uy_fft = self.state_spect.get_var("uy_fft")
            result = self.oper.rotfft_from_vecfft(ux_fft, uy_fft)
        elif key == "div_fft":
            ux_fft = self.state_spect.get_var("ux_fft")
            uy_fft = self.state_spect.get_var("uy_fft")
            result = self.oper.divfft_from_vecfft(ux_fft, uy_fft)
        elif key == "div":
            div_fft = self.compute("div_fft")
            result = self.oper.ifft2(div_fft)
        elif key == "q":
            rot = self.state_phys.get_var("rot")
            eta = self.state_phys.get_var("eta")
            result = rot - self.params.f * eta
        elif key == "q_fft":
            ux_fft = self.state_spect.get_var("ux_fft")
            uy_fft = self.state_spect.get_var("uy_fft")
            eta_fft = self.state_spect.get_var("eta_fft")
            rot_fft = self.oper.rotfft_from_vecfft(ux_fft, uy_fft)
            result = rot_fft - self.params.f * eta_fft

        elif key == "a_fft":
            ux_fft = self.state_spect.get_var("ux_fft")
            uy_fft = self.state_spect.get_var("uy_fft")
            eta_fft = self.state_spect.get_var("eta_fft")
            result = self.oper.afft_from_uxuyetafft(ux_fft, uy_fft, eta_fft)

        elif key == "h":
            eta = self.state_phys.get_var("eta")
            result = 1 + eta

        elif key == "Floc":
            h = self.compute("h")
            ux = self.state_phys.get_var("ux")
            uy = self.state_phys.get_var("uy")
            result = np.sqrt((ux ** 2 + uy ** 2) / (self.sim.params.c2 * h))

        else:
            to_print = 'Do not know how to compute "' + key + '".'
            if RAISE_ERROR:
                raise ValueError(to_print)

            else:
                if mpi.rank == 0:
                    print(to_print + "\nreturn an array of zeros.")

                result = self.oper.create_arrayX(value=0.0)

        if SAVE_IN_DICT:
            self.vars_computed[key] = result
            self.it_computed[key] = it

        return result

    def statespect_from_statephys(self):
        """Compute the state in Fourier space."""
        ux = self.state_phys.get_var("ux")
        uy = self.state_phys.get_var("uy")
        eta = self.state_phys.get_var("eta")

        ux_fft = self.state_spect.get_var("ux_fft")
        uy_fft = self.state_spect.get_var("uy_fft")
        eta_fft = self.state_spect.get_var("eta_fft")

        self.oper.fft_as_arg(ux, ux_fft)
        self.oper.fft_as_arg(uy, uy_fft)
        self.oper.fft_as_arg(eta, eta_fft)

    def statephys_from_statespect(self, state_spect=None, state_phys=None):
        """Compute the state in physical space."""
        ifft_as_arg = self.oper.ifft_as_arg
        if state_spect is None:
            state_spect = self.state_spect

        if state_phys is None:
            state_phys = self.state_phys

        ux_fft = state_spect.get_var("ux_fft")
        uy_fft = state_spect.get_var("uy_fft")
        eta_fft = state_spect.get_var("eta_fft")
        rot_fft = self.oper.rotfft_from_vecfft(ux_fft, uy_fft)

        ux = state_phys.get_var("ux")
        uy = state_phys.get_var("uy")
        eta = state_phys.get_var("eta")
        rot = state_phys.get_var("rot")

        ifft_as_arg(ux_fft, ux)
        ifft_as_arg(uy_fft, uy)
        ifft_as_arg(eta_fft, eta)
        ifft_as_arg(rot_fft, rot)

    def return_statephys_from_statespect(self, state_spect=None):
        """Return the state in physical space as a new object separate from
        ``self.state_phys``.

        """
        state_phys = SetOfVariables(like=self.state_phys)
        self.statephys_from_statespect(state_spect, state_phys)
        return state_phys

    def init_statespect_from(self, **kwargs):
        """Initializes *state_spect* using arrays provided as keyword
        arguments.

        """
        if len(kwargs) == 1:
            key_fft, value = list(kwargs.items())[0]
            try:
                init_from_keyfft = self.__getattribute__(
                    "init_from_" + key_fft.replace("_", "")
                )
                init_from_keyfft(value)
            except AttributeError:
                super().init_statespect_from(**kwargs)
        elif len(kwargs) == 2:
            if "q_fft" in kwargs and "a_fft" in kwargs:
                self.init_from_qafft(**kwargs)
        else:
            super().init_statespect_from(**kwargs)

    def init_from_etafft(self, eta_fft):
        r"""Initialize from :math:`\hat \eta` and set velocities to zero."""
        state_spect = self.state_spect
        state_spect.set_var("ux_fft", np.zeros_like(eta_fft))
        state_spect.set_var("uy_fft", np.zeros_like(eta_fft))
        state_spect.set_var("eta_fft", eta_fft)

        self.oper.dealiasing(state_spect)
        self.statephys_from_statespect()

    def init_from_uxuyetafft(self, ux_fft, uy_fft, eta_fft):
        """Self explanatory."""
        state_spect = self.state_spect
        state_spect.set_var("ux_fft", ux_fft)
        state_spect.set_var("uy_fft", uy_fft)
        state_spect.set_var("eta_fft", eta_fft)

        self.oper.dealiasing(state_spect)
        self.statephys_from_statespect()

    def init_from_rotuxuyfft(self, rot, ux_fft, uy_fft):
        """Initializes from velocities and discards vorticity."""
        self.init_from_uxuyfft(ux_fft, uy_fft)

    def init_from_rotfft(self, rot_fft):
        """Initializes with rotational velocities computed from vorticity."""
        ux_fft, uy_fft = self.oper.vecfft_from_rotfft(rot_fft)
        self.init_from_uxuyfft(ux_fft, uy_fft)

    def init_from_qfft(self, q_fft):
        """Initialize from potential vorticity."""
        rot_fft = self.oper.rotfft_from_qfft(q_fft)
        ux_fft, uy_fft = self.oper.vecfft_from_rotfft(rot_fft)
        eta_fft = self.oper.etafft_from_qfft(q_fft)
        self.init_from_uxuyetafft(ux_fft, uy_fft, eta_fft)

    def init_from_afft(self, a_fft):
        """Initialize from ageostrophic variable."""
        ux_fft, uy_fft, eta_fft = self.oper.uxuyetafft_from_afft(a_fft)
        self.init_from_uxuyetafft(ux_fft, uy_fft, eta_fft)

    def init_from_qafft(self, q_fft, a_fft):
        """Initialize from potential vorticity and ageostrophic variables."""
        rot_fft = self.oper.rotfft_from_qfft(q_fft)
        uxq_fft, uyq_fft = self.oper.vecfft_from_rotfft(rot_fft)
        etaq_fft = self.oper.etafft_from_qfft(q_fft)

        uxa_fft, uya_fft, etaa_fft = self.oper.uxuyetafft_from_afft(a_fft)
        self.init_from_uxuyetafft(
            uxq_fft + uxa_fft, uyq_fft + uxa_fft, etaq_fft + etaa_fft
        )

    def init_from_uxuyfft(self, ux_fft, uy_fft):
        """Initialize from velocities and adjust surface displacement by solving
        a Poisson equation, assuming no mean flow.

        """
        oper = self.oper
        ifft2 = oper.ifft2

        oper.projection_perp(ux_fft, uy_fft)
        oper.dealiasing(ux_fft, uy_fft)

        ux = ifft2(ux_fft)
        uy = ifft2(uy_fft)

        rot_fft = oper.rotfft_from_vecfft(ux_fft, uy_fft)
        rot = ifft2(rot_fft)

        eta_fft = self._etafft_no_div(ux, uy, rot)
        eta = ifft2(eta_fft)

        state_spect = self.state_spect
        state_spect.set_var("ux_fft", ux_fft)
        state_spect.set_var("uy_fft", uy_fft)
        state_spect.set_var("eta_fft", eta_fft)

        state_phys = self.state_phys
        state_phys.set_var("rot", rot)
        state_phys.set_var("ux", ux)
        state_phys.set_var("uy", uy)
        state_phys.set_var("eta", eta)

    def _etafft_no_div(self, ux, uy, rot):
        K2_not0 = self.oper.K2_not0
        rot_abs = rot + self.params.f

        tempx_fft = -self.oper.fft2(rot_abs * uy)
        tempy_fft = self.oper.fft2(rot_abs * ux)

        uu2_fft = self.oper.fft2(ux ** 2 + uy ** 2)

        eta_fft = (
            1.0j * self.oper.KX * tempx_fft / K2_not0
            + 1.0j * self.oper.KY * tempy_fft / K2_not0
            - uu2_fft / 2.0
        ) / self.params.c2

        if mpi.rank == 0:
            eta_fft[0, 0] = 0.0
        self.oper.dealiasing(eta_fft)

        return eta_fft



"""SW1L equations solving exactly the linear terms
==================================================

(:mod:`fluidsim.solvers.sw1l.exactlin.solver`)



"""

import numpy as np


from fluidsim.base.setofvariables import SetOfVariables


from fluidsim.solvers.sw1l.solver import InfoSolverSW1L
from fluidsim.solvers.sw1l.solver import Simul as SimulSW1L


from fluiddyn.util import mpi


class InfoSolverSW1LExactLin(InfoSolverSW1L):
    """Information about the solver SW1L."""

    def _init_root(self):
        super()._init_root()

        sw1l = "fluidsim.solvers.sw1l"

        self.module_name = sw1l + ".exactlin.solver"
        self.class_name = "Simul"
        self.short_name = "SW1Lexlin"

        classes = self.classes

        classes.State.module_name = sw1l + ".exactlin.state"
        classes.State.class_name = "StateSW1LExactLin"


class Simul(SimulSW1L):
    """A solver of the shallow-water 1 layer equations (SW1L)"""

    InfoSolver = InfoSolverSW1LExactLin

    def __init__(self, params):
        if params.beta != 0:
            raise NotImplementedError("Do not use this solver for beta-plane!")

        super().__init__(params)

    def tendencies_nonlin(self, state_spect=None, old=None):
        oper = self.oper
        fft2 = oper.fft2

        if state_spect is None:
            state_phys = self.state.state_phys
            state_spect = self.state.state_spect
        else:
            state_phys = self.state.return_statephys_from_statespect(state_spect)

        ux = state_phys.get_var("ux")
        uy = state_phys.get_var("uy")
        eta = state_phys.get_var("eta")
        rot = state_phys.get_var("rot")

        # compute the nonlinear terms for ux, uy and eta
        N1x = +rot * uy
        N1y = -rot * ux
        gradu2_x_fft, gradu2_y_fft = oper.gradfft_from_fft(
            fft2(ux ** 2 + uy ** 2) / 2
        )

        Nx_fft = fft2(N1x) - gradu2_x_fft
        Ny_fft = fft2(N1y) - gradu2_y_fft

        jx_fft = fft2(eta * ux)
        jy_fft = fft2(eta * uy)
        Neta_fft = -oper.divfft_from_vecfft(jx_fft, jy_fft)

        # self.verify_tendencies(state_spect, state_phys,
        #                        Nx_fft, Ny_fft, Neta_fft)

        # compute the nonlinear terms for q, ap and am
        (Nq_fft, Np_fft, Nm_fft) = self.oper.qapamfft_from_uxuyetafft(
            Nx_fft, Ny_fft, Neta_fft
        )

        # Nq_fft = self.oper.create_arrayK(value=0)
        # Np_fft = self.oper.create_arrayK(value=0)
        # Nm_fft = self.oper.create_arrayK(value=0)

        oper.dealiasing(Nq_fft, Np_fft, Nm_fft)

        if old is None:
            tendencies_fft = SetOfVariables(
                like=self.state.state_spect, info="tendencies_nonlin"
            )
        else:
            tendencies_fft = old
        tendencies_fft.set_var("q_fft", Nq_fft)
        tendencies_fft.set_var("ap_fft", Np_fft)
        tendencies_fft.set_var("am_fft", Nm_fft)

        if self.params.forcing.enable:
            tendencies_fft += self.forcing.get_forcing()

        return tendencies_fft

    def compute_freq_complex(self, key):
        K2 = self.oper.K2
        if key == "q_fft":
            omega = self.oper.create_arrayK(value=0)
        elif key == "ap_fft":
            omega = 1.0j * np.sqrt(self.params.f ** 2 + self.params.c2 * K2)
        elif key == "am_fft":
            omega = -1.0j * np.sqrt(self.params.f ** 2 + self.params.c2 * K2)
        return omega

    # def verify_tendencies(
    #     self, state_spect, state_phys, Nx_fft, Ny_fft, Neta_fft
    # ):
    #     # for verification conservation energy
    #     # compute the linear terms
    #     oper = self.oper
    #     ux = state_phys.get_var("ux")
    #     uy = state_phys.get_var("uy")
    #     eta = state_phys.get_var("eta")

    #     q_fft = state_spect.get_var("q_fft")
    #     ap_fft = state_spect.get_var("ap_fft")
    #     am_fft = state_spect.get_var("am_fft")
    #     a_fft = ap_fft + am_fft
    #     div_fft = oper.divfft_from_apamfft(ap_fft, am_fft)

    #     eta_fft = oper.etafft_from_qfft(q_fft) + oper.etafft_from_afft(a_fft)

    #     dx_c2eta_fft, dy_c2eta_fft = oper.gradfft_from_fft(
    #         self.params.c2 * eta_fft
    #     )
    #     LCx = self.params.f * uy
    #     LCy = -self.params.f * ux
    #     Lx_fft = oper.fft2(LCx) - dx_c2eta_fft
    #     Ly_fft = oper.fft2(LCy) - dy_c2eta_fft
    #     Leta_fft = -div_fft

    #     # compute the full tendencies
    #     Fx_fft = Lx_fft + Nx_fft
    #     Fy_fft = Ly_fft + Ny_fft
    #     Feta_fft = Leta_fft + Neta_fft
    #     oper.dealiasing(Fx_fft, Fy_fft, Feta_fft)

    #     # test : ux, uy, eta ---> q, ap, am
    #     (Fq_fft, Fp_fft, Fm_fft) = self.oper.qapamfft_from_uxuyetafft(
    #         Fx_fft, Fy_fft, Feta_fft
    #     )
    #     # test : q, ap, am ---> ux, uy, eta
    #     (Fx2_fft, Fy2_fft, Feta2_fft) = self.oper.uxuyetafft_from_qapamfft(
    #         Fq_fft, Fp_fft, Fm_fft
    #     )
    #     print(np.max(abs(Fx2_fft - Fx_fft)))
    #     print(np.max(abs(Fy2_fft - Fy_fft)))
    #     print(np.max(abs(Feta2_fft - Feta_fft)))
    #     Fx_fft = Fx2_fft
    #     Fy_fft = Fy2_fft
    #     Feta_fft = Feta2_fft

    #     (Fq2_fft, Fp2_fft, Fm2_fft) = self.oper.qapamfft_from_uxuyetafft(
    #         Fx2_fft, Fy2_fft, Feta2_fft
    #     )
    #     print(np.max(abs(Fq2_fft - Fq_fft)))
    #     print(np.max(abs(Fp2_fft - Fp_fft)))
    #     print(np.max(abs(Fm2_fft - Fm_fft)))

    #     Fx = oper.ifft2(Fx_fft)
    #     Fy = oper.ifft2(Fy_fft)
    #     Feta = oper.ifft2(Feta_fft)
    #     A = (
    #         Feta * (ux ** 2 + uy ** 2) / 2
    #         + (1 + eta) * (ux * Fx + uy * Fy)
    #         + self.params.c2 * eta * Feta
    #     )
    #     A_fft = oper.fft2(A)
    #     if mpi.rank == 0:
    #         print("should be zero =", A_fft[0, 0])


if __name__ == "__main__":

    import fluiddyn as fld

    params = Simul.create_default_params()

    params.short_name_type_run = "test"

    nh = 64
    Lh = 2 * np.pi
    params.oper.nx = nh
    params.oper.ny = nh
    params.oper.Lx = Lh
    params.oper.Ly = Lh

    delta_x = params.oper.Lx / params.oper.nx
    params.nu_8 = (
        2.0 * 10e-1 * params.forcing.forcing_rate ** (1.0 / 3) * delta_x ** 8
    )

    params.time_stepping.t_end = 2.0

    params.init_fields.type = "noise"

    params.forcing.enable = True
    params.forcing.type = "waves"

    params.output.periods_print.print_stdout = 0.25

    params.output.periods_save.phys_fields = 1.0
    params.output.periods_save.spectra = 0.5
    params.output.periods_save.spect_energy_budg = 0.5
    params.output.periods_save.increments = 0.5
    params.output.periods_save.pdf = 0.5
    params.output.periods_save.time_signals_fft = False

    params.output.periods_plot.phys_fields = 0.0

    params.output.phys_fields.field_to_plot = "div"

    sim = Simul(params)

    # sim.output.phys_fields.plot()
    sim.time_stepping.start()
    # sim.output.phys_fields.plot()

    fld.show()



"""Basilisk solvers (:mod:`fluidsim.base.basilisk`)
===================================================
A reference implementation to use 2D solvers in `Basilisk <http://basilisk.fr>`__ via
FluidSim.

Provides:

.. autosummary::
   :toctree:

    solver
    state
    operators
    time_stepping
    output

"""



from fluidsim.base.solvers.base import InfoSolverBase


class InfoSolverFiniteDiff(InfoSolverBase):
    """Contain the information on a solver."""

    def _init_root(self):

        super()._init_root()

        # self.classes.State.module_name = 'fluidsim.base.state'
        # self.classes.State.class_name = 'StateBase'

        self.classes.TimeStepping.module_name = (
            "fluidsim.base.time_stepping.finite_diff"
        )
        self.classes.TimeStepping.class_name = (
            "TimeSteppingFiniteDiffCrankNicolson"
        )

        self.classes.Operators.module_name = "fluidsim.operators.op_finitediff1d"
        self.classes.Operators.class_name = "OperatorFiniteDiff1DPeriodic"


# # TODO: When required write a PreprocessFiniteDiff class.
# self.classes._set_child(
#     'Preprocess',
#     attribs={'module_name':
#              'fluidsim.base.preprocess.base',
#              'class_name': 'PreprocessBase'})



"""
Spatiotemporal Spectra (:mod:`fluidsim.solvers.ns3d.output.spatiotemporal_spectra`)
===================================================================================

Provides:

.. autoclass:: SpatioTemporalSpectraNS3D
   :members:
   :private-members:

"""

from math import pi

import numpy as np

from fluidsim.base.output.spatiotemporal_spectra import (
    SpatioTemporalSpectra3D,
    SpatioTemporalSpectraNS,
)

from transonic import boost, Array, Type

A4 = Array[Type(np.float32, np.float64), "4d", "C"]
A3 = "float[:,:,:]"
A1 = "float[:]"


@boost
def compute_spectrum_kzkhomega(
    field_k0k1k2omega: A4, khs: A1, kzs: A1, KX: A3, KZ: A3, KH: A3
):
    """Compute the kz-kh-omega spectrum."""
    deltakh = khs[1]
    deltakz = kzs[1]

    nkh = len(khs)
    nkz = len(kzs)
    nk0, nk1, nk2, nomega = field_k0k1k2omega.shape
    spectrum_kzkhomega = np.zeros(
        (nkz, nkh, nomega), dtype=field_k0k1k2omega.dtype
    )

    for ik0 in range(nk0):
        for ik1 in range(nk1):
            for ik2 in range(nk2):
                values = field_k0k1k2omega[ik0, ik1, ik2, :]
                kx = KX[ik0, ik1, ik2]
                if kx != 0.0:
                    # warning: we should also consider another condition
                    # (kx != kx_max) but it is not necessary here mainly
                    # because of dealiasing
                    values = 2 * values

                kappa = KH[ik0, ik1, ik2]
                ikh = int(kappa / deltakh)
                kz = abs(KZ[ik0, ik1, ik2])
                ikz = int(round(kz / deltakz))
                if ikz >= nkz - 1:
                    ikz = nkz - 1
                if ikh >= nkh - 1:
                    ikh = nkh - 1
                    for i, value in enumerate(values):
                        spectrum_kzkhomega[ikz, ikh, i] += value
                else:
                    coef_share = (kappa - khs[ikh]) / deltakh
                    for i, value in enumerate(values):
                        spectrum_kzkhomega[ikz, ikh, i] += (
                            1 - coef_share
                        ) * value
                        spectrum_kzkhomega[ikz, ikh + 1, i] += coef_share * value

    # get one-sided spectrum in the omega dimension
    nomega = (nomega + 1) // 2
    spectrum_onesided = np.zeros((nkz, nkh, nomega))
    spectrum_onesided[:, :, 0] = spectrum_kzkhomega[:, :, 0]
    spectrum_onesided[:, :, 1:] = (
        spectrum_kzkhomega[:, :, 1:nomega]
        + spectrum_kzkhomega[:, :, -1:-nomega:-1]
    )
    return spectrum_onesided / (deltakz * deltakh)


def _sum_wavenumber3D(field, KX, kx_max):
    n0, n1, n2 = field.shape[:3]
    result = 0.0
    for i0 in range(n0):
        for i1 in range(n1):
            for i2 in range(n2):
                value = field[i0, i1, i2]
                kx = KX[i0, i1, i2]
                if kx != 0.0 and kx != kx_max:
                    value *= 2
                result += value
    return result


class SpatioTemporalSpectraNS3D(SpatioTemporalSpectraNS, SpatioTemporalSpectra3D):

    compute_spectrum_kzkhomega = staticmethod(compute_spectrum_kzkhomega)
    _sum_wavenumber = staticmethod(_sum_wavenumber3D)

    def compute_spectra_urud(self, tmin=0, tmax=None, dtype=None):
        """compute the spectra of ur, ud from files"""
        if tmax is None:
            tmax = self.sim.params.time_stepping.t_end

        # load time series as state_spect arrays + times
        series = self.load_time_series(
            keys=("vx", "vy"), tmin=tmin, tmax=tmax, dtype=dtype
        )

        # toroidal/poloidal decomposition
        # urx_fft, ury_fft contain shear modes!
        vx_fft = series["vx_Fourier"]
        vy_fft = series["vy_Fourier"]
        if vx_fft.dtype == "complex64":
            float_dtype = "float32"
        elif vx_fft.dtype == "complex128":
            float_dtype = "float64"

        params_oper = self.sim.params.oper
        deltaky = 2 * pi / params_oper.Ly
        deltakx = 2 * pi / params_oper.Lx

        order = series["dims_order"]

        shapeK = series[f"K{order[1]}_adim"].shape
        KY = np.zeros(shapeK + (1,), dtype=float_dtype)
        KX = np.zeros(shapeK + (1,), dtype=float_dtype)
        KY[..., 0] = deltaky * series[f"K{order[1]}_adim"]
        KX[..., 0] = deltakx * series[f"K{order[2]}_adim"]

        inv_Kh_square_nozero = KX ** 2 + KY ** 2
        inv_Kh_square_nozero[inv_Kh_square_nozero == 0] = 1e-14
        inv_Kh_square_nozero = 1 / inv_Kh_square_nozero

        kdotu_fft = KX * vx_fft + KY * vy_fft
        udx_fft = kdotu_fft * KX * inv_Kh_square_nozero
        udy_fft = kdotu_fft * KY * inv_Kh_square_nozero

        urx_fft = vx_fft - udx_fft
        ury_fft = vy_fft - udy_fft

        del vx_fft, vy_fft, KX, KY, inv_Kh_square_nozero, kdotu_fft

        # perform time fft
        print("computing temporal spectra...")

        spectra = {k: v for k, v in series.items() if k.startswith("K")}
        del series

        # ud
        spectra["spectrum_Khd"] = Khd = np.zeros(udx_fft.shape, dtype=dtype)
        freq, spectrum = self._compute_spectrum(udx_fft)
        Khd += 0.5 * spectrum
        freq, spectrum = self._compute_spectrum(udy_fft)
        Khd += 0.5 * spectrum

        # ur
        spectra["spectrum_Khr"] = Khr = np.zeros(udx_fft.shape, dtype=dtype)
        freq, spectrum = self._compute_spectrum(urx_fft)
        Khr += 0.5 * spectrum
        freq, spectrum = self._compute_spectrum(ury_fft)
        Khr += 0.5 * spectrum

        spectra["omegas"] = 2 * pi * freq
        spectra["dims_order"] = order

        return spectra



"""Movies output (:mod:`fluidsim.base.output.movies`)
=====================================================

Contains base classes which acts as a framework to
implement the method `animate` to make movies.


Provides:

.. autoclass:: MoviesBase
   :members:
   :private-members:

.. autoclass:: MoviesBase1D
   :members:
   :private-members:

.. autoclass:: MoviesBase2D
   :members:
   :private-members:

"""

import os
import numpy as np
import matplotlib.pyplot as plt
from matplotlib import animation
from fluiddyn.util import mpi, is_run_from_jupyter


class MoviesBase:
    """Base class defining most generic functions for movies."""

    def __init__(self, output):
        params = output.sim.params
        self.output = output
        self.sim = output.sim
        self.params = params.output
        self.oper = self.sim.oper

        self._set_font()

    def init_animation(
        self, key_field, numfig, dt_equations, tmin, tmax, fig_kw, **kwargs
    ):
        """Initializes animated fig. and list of times of save files to load."""
        self._set_key_field(key_field)
        self._init_ani_times(tmin, tmax, dt_equations)
        self.fig, self.ax = plt.subplots(num=numfig, **fig_kw)
        self._init_labels()

    def _init_ani_times(self, tmin, tmax, dt_equations):
        """Initialization of the variable ani_times for one animation."""
        self.phys_fields.set_of_phys_files.update_times()
        if tmax is None:
            tmax = self.sim.time_stepping.t

        if tmin is None:
            tmin = self.phys_fields.set_of_phys_files.get_min_time()

        if tmin > tmax:
            raise ValueError(
                "Error tmin > tmax. Value tmin should be smaller than tmax"
            )

        if dt_equations is None:
            dt_equations = self.params.periods_save.phys_fields

        self.ani_times = np.arange(tmin, tmax + dt_equations, dt_equations)

    def update_animation(self, frame, **fargs):
        """Replace this function to load data for next frame and update the
        figure.

        """
        pass

    def _set_font(self, family="serif", size=12):
        """Use to set font attribute. May be either an alias (generic name
        is CSS parlance), such as serif, sans-serif, cursive, fantasy, or
        monospace, a real font name or a list of real font names.

        """
        self.font = {
            "family": family,
            "color": "black",
            "weight": "normal",
            "size": size,
        }

    def get_field_to_plot(self, time=None, key=None, equation=None):
        """
        Once a saved file is loaded, this selects the field and mpi-gathers.

        Returns
        -------
        field : nd array or string
        """
        raise NotImplementedError(
            "get_field_to_plot function declaration missing."
        )

    def _init_labels(self, xlabel=None, ylabel=None):
        """Initialize the labels."""
        if xlabel is None:
            xlabel = self.sim.oper.axes[1]
        if ylabel is None:
            ylabel = self.sim.oper.axes[0]
        self.ax.set_xlabel(xlabel, fontdict=self.font)
        self.ax.set_ylabel(ylabel, fontdict=self.font)

    def _get_axis_data(self):
        """Replace this function to load axis data."""
        raise NotImplementedError("_get_axis_data  function declaration missing.")

    def _set_key_field(self, key_field):
        """
        Defines key_field default.
        """
        self.key_field = self.phys_fields.get_key_field_to_plot(
            forbid_compute=True, key_field_to_plot=key_field
        )

    def animate(
        self,
        key_field=None,
        dt_frame_in_sec=0.3,
        dt_equations=None,
        tmin=None,
        tmax=None,
        repeat=True,
        save_file=False,
        numfig=None,
        fargs={},
        fig_kw={},
        **kwargs,
    ):
        """Load the key field from multiple save files and display as
        an animated plot or save as a movie file.

        Parameters
        ----------
        key_field : str
            Specifies which field to animate
        dt_frame_in_sec : float
            Interval between animated frames in seconds
        dt_equations : float
            Approx. interval between saved files to load in simulation time
            units
        tmax : float
            Animate till time `tmax`.
        repeat : bool
            Loop the animation
        save_file : str or bool
            Path to save the movie. When `True`  saves into a file instead
            of plotting it on screen (default: ~/fluidsim_movie.mp4). Specify
            a string to save to another file location. Format is autodetected
            from the filename extension.
        numfig : int
            Figure number on the window
        fargs : dict
            Dictionary of arguments for `update_animation`. Matplotlib
            requirement.
        fig_kw : dict
            Dictionary of arguments for arguments for the figure.

        Other Parameters
        ----------------
        All `kwargs` are passed on to `init_animation` and `_ani_save`

        xmax : float
            Set x-axis limit for 1D animated plots
        ymax : float
            Set y-axis limit for 1D animated plots
        clim : tuple
            Set colorbar limits for 2D animated plots
        step : int
            Set step value to get a coarse 2D field
        QUIVER : bool
            Set quiver on or off on top of 2D pcolor plots
        codec : str
            Codec used to save into a movie file (default: ffmpeg)

        Examples
        --------
        >>> import fluidsim as fls
        >>> sim = fls.load_sim_for_plot()
        >>> animate = sim.output.spectra.animate
        >>> animate('E')
        >>> animate('rot')
        >>> animate('rot', dt_equations=0.1, dt_frame_in_sec=50, clim=(-5, 5))
        >>> animate('rot', clim=(-300, 300), fig_kw={"figsize": (14, 4)})
        >>> animate('rot', tmax=25, clim=(-5, 5), save_file='True')
        >>> animate('rot', clim=(-5, 5), save_file='~/fluidsim.gif', codec='imagemagick')

        .. TODO: Use FuncAnimation with blit=True option.

        Notes
        -----

        This method is kept as generic as possible. Any arguments specific to
        1D, 2D or 3D animations or specific to a type of output are be passed
        via keyword arguments (``kwargs``) into its respective
        ``init_animation`` or ``update_animation`` methods.

        """
        if mpi.rank > 0:
            raise NotImplementedError("Do NOT use this function with MPI !")

        self.init_animation(
            key_field, numfig, dt_equations, tmin, tmax, fig_kw, **kwargs
        )

        if isinstance(repeat, int):
            nb_repeat = repeat
            repeat = False
        else:
            nb_repeat = 1

        self._animation = animation.FuncAnimation(
            self.fig,
            self.update_animation,
            nb_repeat * len(self.ani_times),
            fargs=fargs.items(),
            interval=dt_frame_in_sec * 1000,
            blit=False,
            repeat=repeat,
        )

        if save_file:
            if not isinstance(save_file, str):
                save_file = r"~/fluidsim_movie.mp4"

            self._ani_save(save_file, dt_frame_in_sec, **kwargs)

    def interact(
        self,
        key_field=None,
        dt_frame_in_sec=0.3,
        dt_equations=None,
        tmin=None,
        tmax=None,
        fig_kw={},
        **kwargs,
    ):
        """Launches an interactive plot.

        Parameters
        ----------
        key_field : str
            Specifies which field to animate
        dt_frame_in_sec : float
            Interval between animated frames in seconds
        dt_equations : float
            Approx. interval between saved files to load in simulation time
            units
        tmax : float
            Animate till time `tmax`.
        fig_kw : dict
            Dictionary of arguments for arguments for the figure.

        Other Parameters
        ----------------
        All `kwargs` are passed on to `init_animation` and `_ani_save`

        xmax : float
            Set x-axis limit for 1D animated plots
        ymax : float
            Set y-axis limit for 1D animated plots
        clim : tuple
            Set colorbar limits for 2D animated plots
        step : int
            Set step value to get a coarse 2D field
        QUIVER : bool
            Set quiver on or off on top of 2D pcolor plots

        Notes
        -----
        Installation instructions for notebook::

          pip install ipywidgets
          jupyter nbextension enable --py widgetsnbextension

        Restart the notebook and call the function using::

          >>> %matplotlib notebook

        For JupyterLab::

          pip install ipywidgets ipympl
          jupyter labextension install @jupyter-widgets/jupyterlab-manager

        Restart JupyterLab and call the function using::

          >>> %matplotlib widget

        """
        try:
            from ipywidgets import interact, widgets
        except ImportError:
            raise ImportError(
                "See fluidsim.base.output.movies.interact docstring."
            )

        if not is_run_from_jupyter():
            raise ValueError("Works only inside Jupyter.")

        self.init_animation(
            key_field, 0, dt_equations, tmin, tmax, fig_kw, **kwargs
        )
        if tmin is None:
            tmin = self.ani_times[0]
        if tmax is None:
            tmax = self.ani_times[-1]
        if dt_equations is None:
            dt_equations = self.ani_times[1] - self.ani_times[0]

        slider = widgets.FloatSlider(
            min=float(tmin),
            max=float(tmax),
            step=float(dt_equations),
            value=float(tmin),
        )

        def widget_update(time):
            frame = np.argmin(abs(self.ani_times - time))
            self.update_animation(frame)
            self.fig.canvas.draw()

        interact(widget_update, time=slider)

    def _ani_save(self, path_file, dt_frame_in_sec, codec="ffmpeg", **kwargs):
        """Saves the animation using `matplotlib.animation.writers`."""

        path_file = os.path.expandvars(path_file)
        path_file = os.path.expanduser(path_file)
        avail = animation.writers.list()
        if len(avail) == 0:
            raise ValueError(
                "Please install a codec library. For e.g. ffmpeg, mencoder, "
                "imagemagick, html"
            )

        elif codec not in avail:
            print("Using one of the available codecs: {}".format(avail.keys()))
            codec = list(avail.keys())[0]

        Writer = animation.writers[codec]

        print("Saving movie to ", path_file, "...")
        writer = Writer(
            fps=1.0 / dt_frame_in_sec, metadata=dict(artist="FluidSim")
        )
        # _animation is a FuncAnimation object
        self._animation.save(path_file, writer=writer, dpi=150)


class MoviesBase1D(MoviesBase):
    """Base class defining most generic functions for movies for 1D data."""

    def init_animation(
        self, key_field, numfig, dt_equations, tmin, tmax, fig_kw, **kwargs
    ):
        """Initializes animated figure."""

        super().init_animation(
            key_field, numfig, dt_equations, tmin, tmax, fig_kw, **kwargs
        )

        ax = self.ax
        (self._ani_line,) = ax.plot([], [])

        if "xmax" in kwargs:
            ax.set_xlim(0, kwargs["xmax"])
        else:
            ax.set_xlim(0, self.output.sim.oper.lx)

        if "ymax" in kwargs:
            ax.set_ylim(1e-16, kwargs["ymax"])

    def update_animation(self, frame, **fargs):
        """Loads contour data and updates figure."""
        print("update_animation for frame", frame, "       ", end="\r")
        time = self.ani_times[frame % len(self.ani_times)]
        get_field_to_plot = self.phys_fields.get_field_to_plot
        y, time = get_field_to_plot(time=time, key=self.key_field)
        x = self._get_axis_data()

        self._ani_line.set_data(x, y)
        self.ax.set_title(
            self.key_field + f", $t = {time:.3f}$\n" + self.output.summary_simul
        )

        return self._ani_line

    def _get_axis_data(self):
        """Get axis data.

        Returns
        -------

        x : array
          x-axis data.

        """
        try:
            x = self.oper.xs
        except AttributeError:
            x = self.oper.x_seq
        return x


class MoviesBase2D(MoviesBase):
    """Base class defining most generic functions for movies for 2D data."""

    def _get_axis_data(self):
        """Get axis data.

        Returns
        -------

        x : array
          x-axis data.

        y : array
          y-axis data.

        """

        if not hasattr(self, "_equation"):
            equation = None
        else:
            equation = self._equation

        if (
            equation is None
            or equation.startswith("iz=")
            or equation.startswith("z=")
        ):
            x = self.oper.get_grid1d_seq("x")
            y = self.oper.get_grid1d_seq("y")
        elif equation.startswith("iy=") or equation.startswith("y="):
            x = self.oper.get_grid1d_seq("x")
            y = self.oper.get_grid1d_seq("z")
        elif equation.startswith("ix=") or equation.startswith("x="):
            x = self.oper.get_grid1d_seq("y")
            y = self.oper.get_grid1d_seq("z")
        else:
            raise NotImplementedError

        return x, y



# -*- coding: utf-8 -*-

"""Plate2d solver (:mod:`fluidsim.solvers.plate2d.dimensional`)
================================================================

Provides:

.. autoclass:: Converter
   :members:
   :private-members:

"""

import numpy as np


class Converter:
    r"""Converter between dimensional and adimensional variables.

    Parameters
    ----------

    C : number
        Defined by the dispersion relation :math:`\omega = C k^2`.

    h : number
        Thickness of the plate (in meter).

    L : number, optional
        Length of the plate (in meter).

    sigma : number, optional
        Poisson modulus.

    Notes
    -----

    .. math::
       C = \frac{E h^2}{12\rho(1 - \sigma^2)}

    """

    def __init__(self, C, h, L=1.0, sigma=0.0):
        self.C = C
        self.h = h
        self.L = L
        self.sigma = sigma

        self.tilde_T = L ** 2 / np.sqrt(C)
        self.tilde_Z = h / np.sqrt(6 * (1 - sigma))
        self.tilde_L = L
        self.tilde_nu4 = L ** 4 / self.tilde_T

    def compute_time_adim(self, t):
        """Compute the dimensional time."""
        return t / self.tilde_T

    def compute_time_dim(self, t):
        """Compute the adimensional time."""
        return t * self.tilde_T

    def compute_z_adim(self, z):
        """Compute the dimensional deformation."""
        return z / self.tilde_Z

    def compute_z_dim(self, z):
        """Compute the adimensional deformation."""
        return z * self.tilde_Z

    def compute_w_adim(self, w):
        """Compute the dimensional deformation."""
        return self.tilde_T / self.tilde_Z * w

    def compute_w_dim(self, w):
        """Compute the adimensional deformation."""
        return self.tilde_Z / self.tilde_T * w

    def compute_nu4_adim(self, nu4):
        """Compute the dimensional hyper-viscosity."""
        return nu4 / self.tilde_nu4

    def compute_nu4_dim(self, nu4):
        """Compute the adimensional hyper-viscosity."""
        return nu4 * self.tilde_nu4


if __name__ == "__main__":

    converter = Converter(C=0.648 ** 2, h=4e-4)

    amplitude_z = 0.001
    print("amplitude z: {:6.2f}".format(converter.compute_z_adim(amplitude_z)))

    nu_4 = 5e-7
    print("nu_4: {:8.3g}".format(converter.compute_nu4_adim(nu_4)))



"""1-layer Shallow-Water solvers (:mod:`fluidsim.solvers.sw1l`)
=====================================================================

Provides

.. autosummary::
   :toctree:

   solver
   state
   operators
   forcing
   init_fields
   output

Modyfied solvers

.. autosummary::
   :toctree:

   exactlin
   modified
   onlywaves

"""



import unittest
from numpy.testing import assert_array_almost_equal
import matplotlib.pyplot as plt
import fluiddyn.util.mpi as mpi
from fluidsim.util.testing import (
    TestSimulConserve,
    skip_if_no_fluidfft,
    classproperty,
)


@skip_if_no_fluidfft
class TestSimulSW1LWaves(TestSimulConserve):
    zero = 1e-6

    @classproperty
    def Simul(cls):
        from fluidsim.solvers.sw1l.onlywaves.solver import Simul

        return Simul

    @classmethod
    def init_params(cls):
        super().init_params()
        params = cls.params
        params.output.HAS_TO_SAVE = True

        params.output.periods_save.spect_energy_budg = 0.2
        params.init_fields.type = "noise"
        params.oper.nx = 16
        params.oper.ny = 8

    def get_tendencies(self):
        tendencies_fft = self.tendencies_fft

        Fq_fft = self.sim.oper.create_arrayK(value=0.0j)
        Fap_fft = tendencies_fft.get_var("ap_fft")
        Fam_fft = tendencies_fft.get_var("am_fft")

        return self.sim.oper.uxuyetafft_from_qapamfft(Fq_fft, Fap_fft, Fam_fft)

    def test_energy_conservation(self):
        """Verify that the energy growth rate due to nonlinear tendencies
        (advection term) must be zero.

        """
        Fx_fft, Fy_fft, Feta_fft = self.get_tendencies()
        state_phys = self.sim.state.state_phys
        ux = state_phys.get_var("ux")
        uy = state_phys.get_var("uy")
        eta = state_phys.get_var("eta")

        oper = self.sim.oper
        Fx = oper.ifft2(Fx_fft)
        Fy = oper.ifft2(Fy_fft)
        Feta = oper.ifft2(Feta_fft)
        A = (
            Feta * (ux ** 2 + uy ** 2) / 2
            + (1 + eta) * (ux * Fx + uy * Fy)
            + self.sim.params.c2 * eta * Feta
        )

        A_fft = oper.fft2(A)
        if mpi.rank == 0:
            self.assertAlmostZero(A_fft[0, 0], tolerance_warning=False)

    def test_state_methods(self):
        state_spect = self.sim.state.state_spect.copy()
        self.sim.state.statespect_from_statephys()
        state_spect2 = self.sim.state.state_spect
        assert id(state_spect) != id(state_spect2)
        assert_array_almost_equal(state_spect, state_spect2)

    def test_init_methods(self):
        compute = self.sim.state.compute
        ux_fft, uy_fft, eta_fft = map(compute, ("ux_fft", "uy_fft", "eta_fft"))
        self.sim.state.init_from_uxuyetafft(ux_fft, uy_fft, eta_fft)
        self.sim.state.init_from_etafft(eta_fft)

    def test_state_compute(self):
        for key in ("uy_fft", "rot", "q"):
            var_computed = self.sim.state.compute(key)


if __name__ == "__main__":
    unittest.main()



"""Spectra (:mod:`fluidsim.solvers.plate2d.output.spectra`)
=================================================================


Provides:

.. autoclass:: SpectraPlate2D
   :members:
   :private-members:

"""

import h5py

import numpy as np

from fluidsim.base.output.spectra import Spectra


class SpectraPlate2D(Spectra):
    """Compute, save, load and plot spectra."""

    def compute(self):
        """compute the values at one time."""
        EK_fft, EL_fft, EE_fft = self.output.compute_energies_fft()
        # compute the spectra 1D
        spectrum1Dkx_EK, spectrum1Dky_EK = self.spectra1D_from_fft(EK_fft)
        spectrum1Dkx_EL, spectrum1Dky_EL = self.spectra1D_from_fft(EL_fft)
        spectrum1Dkx_EE, spectrum1Dky_EE = self.spectra1D_from_fft(EE_fft)

        dict_spectra1D = {
            "spectrum1Dkx_EK": spectrum1Dkx_EK,
            "spectrum1Dky_EK": spectrum1Dky_EK,
            "spectrum1Dkx_EL": spectrum1Dkx_EL,
            "spectrum1Dky_EL": spectrum1Dky_EL,
            "spectrum1Dkx_EE": spectrum1Dkx_EE,
            "spectrum1Dky_EE": spectrum1Dky_EE,
        }

        # compute the spectra 2D
        spectrum2D_EK = self.spectrum2D_from_fft(EK_fft)
        spectrum2D_EL = self.spectrum2D_from_fft(EL_fft)
        spectrum2D_EE = self.spectrum2D_from_fft(EE_fft)
        dict_spectra2D = {
            "spectrum2D_EK": spectrum2D_EK,
            "spectrum2D_EL": spectrum2D_EL,
            "spectrum2D_EE": spectrum2D_EE,
        }
        return dict_spectra1D, dict_spectra2D

    def _online_plot_saving(self, dict_spectra1D, dict_spectra2D):
        if (
            self.nx == self.params.oper.ny
            and self.params.oper.Lx == self.params.oper.Ly
        ):
            spectrum2D_EK = dict_spectra2D["spectrum2D_EK"]
            spectrum2D_EL = dict_spectra2D["spectrum2D_EL"]
            spectrum2D_EE = dict_spectra2D["spectrum2D_EE"]
            spectrum2D_Etot = spectrum2D_EK + spectrum2D_EL + spectrum2D_EE
            khE = self.oper.khE
            coef_norm = khE ** (3.0)
            self.axe.loglog(khE, spectrum2D_Etot * coef_norm, "k", linewidth=2)
            self.axe.loglog(khE, spectrum2D_EK * coef_norm, "r--")
            self.axe.loglog(khE, spectrum2D_EL * coef_norm, "b--")
            self.axe.loglog(khE, spectrum2D_EE * coef_norm, "y--")
        # lin_inf, lin_sup = self.axe.get_ylim()
        # if lin_inf < 10e-6:
        #     lin_inf = 10e-6
        # self.axe.set_ylim([lin_inf, lin_sup])
        else:
            print(
                "you need to implement the ploting "
                "of the spectra for this case"
            )

    def plot1d(self, tmin=0, tmax=1000, delta_t=2, coef_compensate=3):

        with h5py.File(self.path_file1D, "r") as h5file:
            dset_times = h5file["times"]

            dset_kxE = h5file["kxE"]
            # dset_kyE = h5file['kyE']
            kh = dset_kxE[...]

            kh_not0 = kh.copy()
            kh_not0[kh_not0 == 0] = 1e-15

            dset_spectrum1Dkx_EK = h5file["spectrum1Dkx_EK"]
            dset_spectrum1Dky_EK = h5file["spectrum1Dky_EK"]
            # dset_spectrum1Dkx_EL = h5file['spectrum1Dkx_EL']
            # dset_spectrum1Dky_EL = h5file['spectrum1Dky_EL']
            # dset_spectrum1Dkx_EE = h5file['spectrum1Dkx_EE']
            # dset_spectrum1Dky_EE = h5file['spectrum1Dky_EE']

            times = dset_times[...]

            delta_t_save = np.mean(times[1:] - times[0:-1])
            delta_i_plot = int(np.round(delta_t / delta_t_save))
            delta_t = delta_t_save * delta_i_plot
            if delta_i_plot == 0:
                delta_i_plot = 1

            imin_plot = np.argmin(abs(times - tmin))
            imax_plot = np.argmin(abs(times - tmax))

            tmin_plot = times[imin_plot]
            tmax_plot = times[imax_plot]

            print(
                "plot1d(tmin={}, tmax={}, delta_t={:.2f},".format(
                    tmin, tmax, delta_t
                )
                + f" coef_compensate={coef_compensate:.3f})"
            )

            print(
                (
                    "plot 1D spectra\n"
                    "tmin = {0:8.6g} ; tmax = {1:8.6g} ; delta_t = {2:8.6g}\n"
                    "imin = {3:8d} ; imax = {4:8d} ; delta_i = {5:8d}"
                ).format(
                    tmin_plot,
                    tmax_plot,
                    delta_t,
                    imin_plot,
                    imax_plot,
                    delta_i_plot,
                )
            )

            fig, ax1 = self.output.figure_axe()
            ax1.set_xlabel("$k_h$")
            ax1.set_ylabel("spectra")
            ax1.set_title("1D spectra\n" + self.output.summary_simul)
            ax1.set_xscale("log")
            ax1.set_yscale("log")

            coef_norm = kh_not0 ** (coef_compensate)
            if delta_t != 0.0:
                for it in range(imin_plot, imax_plot + 1, delta_i_plot):
                    EK = dset_spectrum1Dkx_EK[it] + dset_spectrum1Dky_EK[it]
                    EK[EK < 10e-16] = 0.0
                    ax1.plot(kh, EK * coef_norm, "r", linewidth=1)

            EK = (
                dset_spectrum1Dkx_EK[imin_plot : imax_plot + 1]
                + dset_spectrum1Dky_EK[imin_plot : imax_plot + 1]
            ).mean(0)

        ax1.plot(kh, kh_not0 ** (-3) * coef_norm, "k", linewidth=1)
        ax1.plot(kh, 0.01 * kh_not0 ** (-5 / 3) * coef_norm, "k--", linewidth=1)

    def plot2d(self, tmin=0, tmax=1000, delta_t=2, coef_compensate=3):
        with h5py.File(self.path_file2D, "r") as h5file:
            dset_times = h5file["times"]
            nt = dset_times.shape[0]
            if nt == 0:
                raise ValueError("No spectra are saved in this file.")

            times = dset_times[...]

            kh = h5file["khE"][...]

            kh_not0 = kh.copy()
            kh_not0[kh_not0 == 0] = 1e-15

            dset_spectrum_EK = h5file["spectrum2D_EK"]
            dset_spectrum_EL = h5file["spectrum2D_EL"]
            dset_spectrum_EE = h5file["spectrum2D_EE"]

            if nt == 1:
                imin_plot = imax_plot = 0
            else:
                delta_t_save = np.mean(times[1:] - times[0:-1])
                delta_i_plot = int(np.round(delta_t / delta_t_save))
                if delta_i_plot == 0 and delta_t != 0.0:
                    delta_i_plot = 1
                delta_t = delta_i_plot * delta_t_save

                imin_plot = np.argmin(abs(times - tmin))
                imax_plot = np.argmin(abs(times - tmax))

            tmin_plot = times[imin_plot]
            tmax_plot = times[imax_plot]

            print(
                "plot2d(tmin={}, tmax={}, delta_t={:.2f},".format(
                    tmin, tmax, delta_t
                )
                + f" coef_compensate={coef_compensate:.3f})"
            )

            print(
                (
                    "plot 2D spectra\n"
                    "tmin = {0:8.6g} ; tmax = {1:8.6g} ; delta_t = {2:8.6g}\n"
                    "imin = {3:8d} ; imax = {4:8d} ; delta_i = {5:8d}"
                ).format(
                    tmin_plot,
                    tmax_plot,
                    delta_t,
                    imin_plot,
                    imax_plot,
                    delta_i_plot,
                )
            )

            fig, ax1 = self.output.figure_axe()
            ax1.set_xlabel("$k_h$")
            ax1.set_ylabel("2D spectra")
            ax1.set_title("2D spectra\n" + self.output.summary_simul)
            ax1.set_xscale("log")
            ax1.set_yscale("log")

            coef_norm = kh_not0 ** coef_compensate

            if delta_t != 0.0:
                for it in range(imin_plot, imax_plot + 1, delta_i_plot):
                    EK = dset_spectrum_EK[it]
                    EK[EK < 10e-16] = 0.0
                    EL = dset_spectrum_EL[it]
                    EL[EL < 10e-16] = 0.0
                    EE = dset_spectrum_EE[it]
                    EE[EE < 10e-16] = 0.0
                    Etot = EK + EL + EE

                    # print(Etot)

                    ax1.plot(kh, Etot * coef_norm, "k--", linewidth=2)
                    ax1.plot(kh, EK * coef_norm, "r--", linewidth=1)
                    ax1.plot(kh, EL * coef_norm, "b--", linewidth=1)
                    ax1.plot(kh, EE * coef_norm, "y--", linewidth=1)

            EK = dset_spectrum_EK[imin_plot : imax_plot + 1].mean(0)
            EK[EK < 10e-16] = 0.0
            EL = dset_spectrum_EL[imin_plot : imax_plot + 1].mean(0)
            EL[EK < 10e-16] = 0.0
            EE = dset_spectrum_EE[imin_plot : imax_plot + 1].mean(0)
            EE[EK < 10e-16] = 0.0

        ax1.plot(kh, EK * coef_norm, "r-", linewidth=2)
        ax1.plot(kh, EL * coef_norm, "b-", linewidth=2)
        ax1.plot(kh, EE * coef_norm, "y-", linewidth=2)

        ax1.plot(kh, kh_not0 ** (-3) * coef_norm, "k:", linewidth=1)
        ax1.plot(kh, 0.01 * kh_not0 ** (-5.0 / 3) * coef_norm, "k-.", linewidth=1)



"""Base solver (:mod:`fluidsim.base.solvers.info_base`)
=======================================================

.. autoclass:: InfoSolverBase
   :members:
   :private-members:

"""
from copy import deepcopy

from fluidsim_core.info import InfoSolverCore, create_info_simul  # noqa


def _merged_element(el1, el2):
    result = deepcopy(el1)
    result.extend(deepcopy(el2))
    return result


class InfoSolverBase(InfoSolverCore):
    """Contain the information on a solver."""

    def _init_root(self):
        super()._init_root()

        self._set_attribs(
            {
                "module_name": "fluidsim.base.solvers.base",
                "class_name": "SimulBase",
                "short_name": "Base",
            }
        )

        classes = self.classes

        classes._set_child(
            "Operators",
            attribs={
                "module_name": "fluidsim.operators.operators0d",
                "class_name": "Operators0D",
            },
        )

        classes._set_child(
            "State",
            attribs={
                "module_name": "fluidsim.base.state",
                "class_name": "StateBase",
            },
        )

        classes._set_child(
            "TimeStepping",
            attribs={
                "module_name": "fluidsim.base.time_stepping.simple",
                "class_name": "TimeSteppingSimple",
            },
        )

        classes._set_child(
            "InitFields",
            attribs={
                "module_name": "fluidsim.base.init_fields",
                "class_name": "InitFieldsBase",
            },
        )

        classes._set_child(
            "Forcing",
            attribs={
                "module_name": "fluidsim.base.forcing",
                "class_name": "ForcingBase",
            },
        )

        classes._set_child(
            "Output",
            attribs={
                "module_name": "fluidsim.base.output.base",
                "class_name": "OutputBase",
            },
        )



"""Solver based on Spherical Harmonic (:mod:`fluidsim.base.sphericalharmo`)
===========================================================================

Provides:

.. autosummary::
   :toctree:

   solver
   state

"""



"""State class for the SW1L.exactlin solver
(:mod:`fluidsim.solvers.sw1l.exactlin.state`)
===================================================


Provides:

.. autoclass:: StateSW1LExactLin
   :members:
   :private-members:

"""

from fluidsim.base.setofvariables import SetOfVariables

from fluidsim.solvers.sw1l.state import StateSW1L

from fluiddyn.util import mpi


class StateSW1LExactLin(StateSW1L):
    """
    The class :class:`StateSW1Lexlin` contains the variables corresponding
    to the state and handles the access to other fields for the solver
    SW1L.
    """

    @staticmethod
    def _complete_info_solver(info_solver):
        """Complete the ParamContainer info_solver.

        This is a static method!
        """
        info_solver.classes.State._set_attribs(
            {
                "keys_state_spect": ["ap_fft", "am_fft", "q_fft"],
                "keys_state_phys": ["ux", "uy", "eta", "rot"],
                "keys_computable": [],
                "keys_phys_needed": ["ux", "uy", "eta"],
                "keys_linear_eigenmodes": ["q_fft", "a_fft", "d_fft"],
            }
        )

    def compute(self, key, SAVE_IN_DICT=True, RAISE_ERROR=True):
        it = self.sim.time_stepping.it

        if key in self.vars_computed and it == self.it_computed[key]:
            return self.vars_computed[key]

        if key == "div_fft":
            ap_fft = self.state_spect.get_var("ap_fft")
            am_fft = self.state_spect.get_var("am_fft")
            d_fft = self.oper.divfft_from_apamfft(ap_fft, am_fft)
            result = d_fft

        elif key == "a_fft":
            ap_fft = self.state_spect.get_var("ap_fft")
            am_fft = self.state_spect.get_var("am_fft")
            result = ap_fft + am_fft

        elif key == "rot_fft":
            q_fft = self.state_spect.get_var("q_fft")
            a_fft = self.compute("a_fft")
            result = self.oper.rotfft_from_qfft(
                q_fft
            ) + self.oper.rotfft_from_afft(a_fft)

        elif key == "eta_fft":
            q_fft = self.state_spect.get_var("q_fft")
            a_fft = self.compute("a_fft")
            result = self.oper.etafft_from_qfft(
                q_fft
            ) + self.oper.etafft_from_afft(a_fft)

        elif key == "ux_fft":
            rot_fft = self.compute("rot_fft")
            div_fft = self.compute("div_fft")
            urx_fft, ury_fft = self.oper.vecfft_from_rotfft(rot_fft)
            udx_fft, udy_fft = self.oper.vecfft_from_divfft(div_fft)
            ux_fft = urx_fft + udx_fft
            if mpi.rank == 0:
                ap_fft = self.state_spect.get_var("ap_fft")
                ux_fft[0, 0] = ap_fft[0, 0]
            result = ux_fft
            if SAVE_IN_DICT:
                key2 = "uy_fft"
                uy_fft = ury_fft + udy_fft
                if mpi.rank == 0:
                    am_fft = self.state_spect.get_var("am_fft")
                    uy_fft[0, 0] = am_fft[0, 0]

                self.vars_computed[key2] = uy_fft
                self.it_computed[key2] = it

        elif key == "uy_fft":
            rot_fft = self.compute("rot_fft")
            div_fft = self.compute("div_fft")
            urx_fft, ury_fft = self.oper.vecfft_from_rotfft(rot_fft)
            udx_fft, udy_fft = self.oper.vecfft_from_divfft(div_fft)
            uy_fft = ury_fft + udy_fft
            if mpi.rank == 0:
                am_fft = self.state_spect.get_var("am_fft")
                uy_fft[0, 0] = am_fft[0, 0]
            result = uy_fft
            if SAVE_IN_DICT:
                key2 = "ux_fft"
                ux_fft = urx_fft + udx_fft
                if mpi.rank == 0:
                    ap_fft = self.state_spect.get_var("ap_fft")
                    ux_fft[0, 0] = ap_fft[0, 0]
                self.vars_computed[key2] = ux_fft
                self.it_computed[key2] = it

        else:
            result = super().compute(
                key, SAVE_IN_DICT=SAVE_IN_DICT, RAISE_ERROR=RAISE_ERROR
            )
            SAVE_IN_DICT = False

        if SAVE_IN_DICT:
            self.vars_computed[key] = result
            self.it_computed[key] = it

        return result

    def statespect_from_statephys(self):
        """Compute the state in Fourier space."""
        ux = self.state_phys.get_var("ux")
        uy = self.state_phys.get_var("uy")
        eta = self.state_phys.get_var("eta")

        eta_fft = self.oper.fft2(eta)
        ux_fft = self.oper.fft2(ux)
        uy_fft = self.oper.fft2(uy)

        (q_fft, ap_fft, am_fft) = self.oper.qapamfft_from_uxuyetafft(
            ux_fft, uy_fft, eta_fft
        )

        self.state_spect.set_var("q_fft", q_fft)
        self.state_spect.set_var("ap_fft", ap_fft)
        self.state_spect.set_var("am_fft", am_fft)

    def statephys_from_statespect(self):
        """Compute the state in physical space."""
        ifft2 = self.oper.ifft2
        q_fft = self.state_spect.get_var("q_fft")
        ap_fft = self.state_spect.get_var("ap_fft")
        am_fft = self.state_spect.get_var("am_fft")

        (ux_fft, uy_fft, eta_fft) = self.oper.uxuyetafft_from_qapamfft(
            q_fft, ap_fft, am_fft
        )

        rot_fft = q_fft + self.params.f * eta_fft

        self.state_phys.set_var("ux", ifft2(ux_fft))
        self.state_phys.set_var("uy", ifft2(uy_fft))
        self.state_phys.set_var("eta", ifft2(eta_fft))
        self.state_phys.set_var("rot", ifft2(rot_fft))

    def return_statephys_from_statespect(self, state_spect=None):
        """Return the state in physical space."""
        ifft2 = self.oper.ifft2
        if state_spect is None:
            state_spect = self.state_spect

        q_fft = state_spect.get_var("q_fft")
        ap_fft = state_spect.get_var("ap_fft")
        am_fft = state_spect.get_var("am_fft")

        (ux_fft, uy_fft, eta_fft) = self.oper.uxuyetafft_from_qapamfft(
            q_fft, ap_fft, am_fft
        )

        rot_fft = q_fft + self.params.f * eta_fft

        state_phys = SetOfVariables(like=self.state_phys)
        state_phys.set_var("ux", ifft2(ux_fft))
        state_phys.set_var("uy", ifft2(uy_fft))
        state_phys.set_var("eta", ifft2(eta_fft))
        state_phys.set_var("rot", ifft2(rot_fft))
        return state_phys

    def init_from_uxuyetafft(self, ux_fft, uy_fft, eta_fft):
        (q_fft, ap_fft, am_fft) = self.oper.qapamfft_from_uxuyetafft(
            ux_fft, uy_fft, eta_fft
        )

        state_spect = self.state_spect
        state_spect.set_var("q_fft", q_fft)
        state_spect.set_var("ap_fft", ap_fft)
        state_spect.set_var("am_fft", am_fft)

        self.oper.dealiasing(state_spect)
        self.statephys_from_statespect()

    def init_from_etafft(self, eta_fft):
        (q_fft, ap_fft, am_fft) = self.oper.qapamfft_from_etafft(eta_fft)

        state_spect = self.state_spect
        state_spect.set_var("q_fft", q_fft)
        state_spect.set_var("ap_fft", ap_fft)
        state_spect.set_var("am_fft", am_fft)

        self.oper.dealiasing(state_spect)
        self.statephys_from_statespect()

    def init_from_uxuyfft(self, ux_fft, uy_fft):
        oper = self.oper
        ifft2 = oper.ifft2

        oper.projection_perp(ux_fft, uy_fft)
        oper.dealiasing(ux_fft, uy_fft)

        ux = ifft2(ux_fft)
        uy = ifft2(uy_fft)

        rot_fft = oper.rotfft_from_vecfft(ux_fft, uy_fft)
        rot = ifft2(rot_fft)

        eta_fft = self._etafft_no_div(ux, uy, rot)
        eta = ifft2(eta_fft)

        (q_fft, ap_fft, am_fft) = self.oper.qapamfft_from_uxuyetafft(
            ux_fft, uy_fft, eta_fft
        )

        state_spect = self.state_spect
        state_spect.set_var("q_fft", q_fft)
        state_spect.set_var("ap_fft", ap_fft)
        state_spect.set_var("am_fft", am_fft)

        state_phys = self.state_phys
        state_phys.set_var("rot", rot)
        state_phys.set_var("ux", ux)
        state_phys.set_var("uy", uy)
        state_phys.set_var("eta", eta)



"""Basilisk solver (:mod:`fluidsim.base.basilisk.solver`)
=========================================================

Provides:

.. autoclass:: InfoSolverBasilisk
   :members:
   :private-members:

.. autoclass:: SimulBasilisk
   :members:
   :private-members:

"""

from fluidsim.base.solvers.base import SimulBase
from fluidsim.base.solvers.info_base import InfoSolverBase

import basilisk.stream as basilisk


class InfoSolverBasilisk(InfoSolverBase):
    """Contain the information on a Basilisk solver."""

    def _init_root(self):

        super()._init_root()

        mod = "fluidsim.base.basilisk"

        self.module_name = mod + ".solver"
        self.class_name = "SimulBasilisk"
        self.short_name = "basil"

        classes = self.classes

        classes.State.module_name = mod + ".state"
        classes.State.class_name = "StateBasilisk"

        classes.TimeStepping.module_name = mod + ".time_stepping"
        classes.TimeStepping.class_name = "TimeSteppingBasilisk"

        classes.Operators.module_name = mod + ".operators"
        classes.Operators.class_name = "OperatorsBasilisk2D"

        classes.Output.module_name = mod + ".output"
        classes.Output.class_name = "OutputBasilisk"


class SimulBasilisk(SimulBase):
    """A solver for Basilisk."""

    InfoSolver = InfoSolverBasilisk

    def __init__(self, params):
        """Initialize parameters, state fields, and event loop."""
        bas = self.basilisk = basilisk
        super().__init__(params)

        def init(i, t):
            bas.omega.f = bas.noise

        bas.event(init, t=0.0)


Simul = SimulBasilisk

if __name__ == "__main__":

    import matplotlib.pyplot as plt

    params = Simul.create_default_params()

    params.short_name_type_run = "test"

    params.oper.nx = 128

    params.time_stepping.deltat0 = 2.4
    params.output.periods_print.print_stdout = 1e-15

    sim = Simul(params)
    sim.time_stepping.start()

    sim.output.print_stdout.plot_energy()
    plt.show()



"""Navier-Stokes 2D solvers (:mod:`fluidsim.solvers.ns2d`)
==========================================================

Base package for the Navier-Stokes 2D solvers. The main solver defined
in :mod:`fluidsim.solvers.ns2d.solver` is pseudo-spectral.

The modules for this ns2d solvers are:

.. autosummary::
   :toctree:

   solver
   state
   init_fields
   output
   forcing

There are also a packages for other Navier-Stokes 2D solvers:

.. autosummary::
   :toctree:

   strat
   bouss

"""



import h5py

import os
import numpy as np

from fluiddyn.util import mpi
from fluiddyn.calcul import easypyfft

from fluidsim.base.output.base import SpecificOutput


class TimeSignalsK(SpecificOutput):
    """A :class:`TimeSignalK` object handles the saving of time signals
    in spectral space.

    This class uses the particular functions defined by some solvers
    :func:`linear_eigenmode_from_values_1k` and
    :func`omega_from_wavenumber`.
    """

    _tag = "time_signals_fft"
    _name_file = "time_sigK.h5"

    @staticmethod
    def _complete_params_with_default(params):
        tag = "time_signals_fft"

        params.output.periods_save._set_attrib(tag, 0)
        params.output.periods_plot._set_attrib(tag, 0)

        params.output._set_child(
            tag, attribs={"nb_shells_time_sigK": 4, "nb_k_per_shell_time_sigK": 4}
        )

    def __init__(self, output):
        self.output = output
        sim = output.sim
        params = sim.params

        self.params = params
        self.c2 = params.c2
        self.f = params.f
        self.nx = params.oper.nx

        if not params.output.HAS_TO_SAVE:
            params.output.periods_save.time_signals_fft = False

        if params.output.periods_save.time_signals_fft:
            self._init_save(sim)

        super().__init__(
            output,
            period_save=params.output.periods_save.time_signals_fft,
            period_plot=params.output.periods_plot.time_signals_fft,
        )

    def _init_save(self, sim):
        """
        Sets the attribs determining how many wavenumbers are selected from
        each shell, for which the time series is saved.
        """
        params = self.params
        self.nb_shells = params.output.time_signals_fft.nb_shells_time_sigK
        self.nb_k_per_shell = (
            params.output.time_signals_fft.nb_k_per_shell_time_sigK
        )
        self.nb_k_tot = self.nb_shells * self.nb_k_per_shell

        i_shift = 3
        deltalogk = np.log(params.oper.nx / 2 * params.oper.coef_dealiasing) / (
            self.nb_shells + i_shift
        )

        deltak = sim.oper.deltak

        self.kh_shell = deltak * np.exp(
            deltalogk * np.arange(i_shift, self.nb_shells + i_shift)
        )

        self.kh_shell = deltak * np.round(self.kh_shell / deltak)

        for i_s in range(1, self.nb_shells):
            if self.kh_shell[i_s - 1] == self.kh_shell[i_s]:
                self.kh_shell[i_s] += deltak

        # if mpi.rank == 0:
        #     print 'self.kh_shell/deltak'
        #     print self.kh_shell/sim.oper.deltak

        # hypothese dispersion relation only function of the module
        # of the wavenumber ("shells")
        self.omega_shell = self.output.omega_from_wavenumber(self.kh_shell)

        kx_array_ik_approx = np.empty([self.nb_k_tot])
        ky_array_ik_approx = np.empty([self.nb_k_tot])

        delta_angle = np.pi / (self.nb_k_per_shell - 1)
        for ishell, kh_s in enumerate(self.kh_shell):
            angle = -np.pi / 2
            for ikps in range(self.nb_k_per_shell):
                kx_array_ik_approx[
                    ishell * self.nb_shells + ikps
                ] = kh_s * np.cos(angle)
                ky_array_ik_approx[
                    ishell * self.nb_shells + ikps
                ] = kh_s * np.sin(angle)
                angle += delta_angle

        self.ik0_array_ik = np.empty([self.nb_k_tot], dtype=np.int32)
        self.ik1_array_ik = np.empty([self.nb_k_tot], dtype=np.int32)
        if mpi.nb_proc > 1:
            self.rank_array_ik = np.empty([self.nb_k_tot], dtype=np.int32)

        for ik in range(self.nb_k_tot):
            kx_approx = kx_array_ik_approx[ik]
            ky_approx = ky_array_ik_approx[ik]
            rank_ik, ik0, ik1 = sim.oper.where_is_wavenumber(kx_approx, ky_approx)
            if mpi.nb_proc > 1:
                self.rank_array_ik[ik] = rank_ik
            self.ik0_array_ik[ik] = ik0
            self.ik1_array_ik[ik] = ik1

        self.kx_array_ik = np.empty([self.nb_k_tot])
        self.ky_array_ik = np.empty([self.nb_k_tot])

        for ik in range(self.nb_k_tot):
            ik0_ik = self.ik0_array_ik[ik]
            ik1_ik = self.ik1_array_ik[ik]

            if mpi.nb_proc > 1:
                rank_ik = self.rank_array_ik[ik]
            else:
                rank_ik = 0

            if mpi.rank == rank_ik:
                kx_1k = sim.oper.KX[ik0_ik, ik1_ik]
                ky_1k = sim.oper.KY[ik0_ik, ik1_ik]

            if rank_ik != 0:
                if mpi.rank == rank_ik:
                    data = np.array([kx_1k, ky_1k])
                    mpi.comm.Send([data, mpi.MPI.DOUBLE], dest=0, tag=ik)
                elif mpi.rank == 0:
                    data = np.empty([2], np.float64)
                    mpi.comm.Recv([data, mpi.MPI.DOUBLE], source=rank_ik, tag=ik)
                    kx_1k = data[0]
                    ky_1k = data[1]

            if mpi.rank == 0:
                self.kx_array_ik[ik] = kx_1k
                self.ky_array_ik[ik] = ky_1k

        if mpi.rank == 0:
            self.kh_array_ik = np.sqrt(
                self.kx_array_ik ** 2 + self.ky_array_ik ** 2
            )

            self.omega_array_ik = self.output.omega_from_wavenumber(
                self.kh_array_ik
            )

            self.period_save = np.pi / (8 * self.omega_array_ik.max())
        else:
            self.period_save = 0.0

        if mpi.nb_proc > 1:
            self.period_save = mpi.comm.bcast(self.period_save)

    def _init_files(self, arrays_1st_time=None):
        if not os.path.exists(self.path_file):
            dict_results = self.compute()
            if mpi.rank == 0:
                arrays_1st_time = {
                    "kh_shell": self.kh_shell,
                    "omega_shell": self.omega_shell,
                    "kx_array_ik": self.kx_array_ik,
                    "ky_array_ik": self.ky_array_ik,
                    "kh_array_ik": self.kh_array_ik,
                    "omega_array_ik": self.omega_array_ik,
                }
                self._create_file_from_dict_arrays(
                    self.path_file, dict_results, arrays_1st_time
                )

        if mpi.rank == 0:
            self.file = h5py.File(self.path_file, "r+")
            self.file.attrs["nb_shells"] = self.nb_shells
            self.file.attrs["nb_k_per_shell"] = self.nb_k_per_shell
            self.file.attrs["nb_k_tot"] = self.nb_k_tot
            # the file is kept open during all the simulation
            self.nb_saved_times = 1

        self.t_last_save = self.sim.time_stepping.t

    def _online_save(self):
        """Save the values at one time. """
        tsim = self.sim.time_stepping.t
        if tsim - self.t_last_save >= self.period_save:
            self.t_last_save = tsim
            dict_results = self.compute()
            if mpi.rank == 0:
                self._add_dict_arrays_to_open_file(
                    self.file, dict_results, self.nb_saved_times
                )
                self.nb_saved_times += 1

    def compute(self):
        """compute the values at one time."""

        get_var = self.sim.state.get_var
        ux_fft = get_var("ux_fft")
        uy_fft = get_var("uy_fft")
        eta_fft = get_var("eta_fft")

        if mpi.rank == 0:
            q_array_ik = np.empty([self.nb_k_tot], dtype=np.complex128)
            d_array_ik = np.empty([self.nb_k_tot], dtype=np.complex128)
            a_array_ik = np.empty([self.nb_k_tot], dtype=np.complex128)

        for ik in range(self.nb_k_tot):
            ik0_ik = self.ik0_array_ik[ik]
            ik1_ik = self.ik1_array_ik[ik]

            if mpi.rank == 0:
                kx_ik = self.kx_array_ik[ik]
                ky_ik = self.ky_array_ik[ik]

            if mpi.nb_proc > 1:
                rank_ik = self.rank_array_ik[ik]
            else:
                rank_ik = 0

            if mpi.rank == rank_ik:
                ux_1k = ux_fft[ik0_ik, ik1_ik]
                uy_1k = uy_fft[ik0_ik, ik1_ik]
                eta_1k = eta_fft[ik0_ik, ik1_ik]

            if rank_ik != 0:
                if mpi.rank == rank_ik:
                    data = np.array([ux_1k, uy_1k, eta_1k])
                    mpi.comm.Send([data, mpi.MPI.COMPLEX], dest=0, tag=ik)
                elif mpi.rank == 0:
                    data = np.empty([3], np.complex128)
                    mpi.comm.Recv([data, mpi.MPI.COMPLEX], source=rank_ik, tag=ik)
                    ux_1k = data[0]
                    uy_1k = data[1]
                    eta_1k = data[2]

            if mpi.rank == 0:
                q_1k, d_1k, a_1k = self.output.linear_eigenmode_from_values_1k(
                    ux_1k, uy_1k, eta_1k, kx_ik, ky_ik
                )
                q_array_ik[ik] = q_1k
                d_array_ik[ik] = d_1k
                a_array_ik[ik] = a_1k

        if mpi.rank == 0:
            dict_results = {
                "q_array_ik": q_array_ik,
                "d_array_ik": d_array_ik,
                "a_array_ik": a_array_ik,
            }
            return dict_results

    def load(self):

        if not os.path.exists(self.path_file):
            raise ValueError(
                "no file time_sigK.h5 in\n" + self.output.dir_save_run
            )

        with h5py.File(self.path_file, "r+") as file:

            dset_times = file["times"]
            times = dset_times[...]

            dict_results = {}
            dict_results["times"] = times

            dict_results["nb_shells"] = file.attrs["nb_shells"]
            dict_results["nb_k_per_shell"] = file.attrs["nb_k_per_shell"]
            dict_results["nb_k_tot"] = file.attrs["nb_k_tot"]

            keys_1time = [
                "kh_shell",
                "omega_shell",
                "kx_array_ik",
                "ky_array_ik",
                "kh_array_ik",
                "omega_array_ik",
            ]

            for key in keys_1time:
                dset_temp = file[key]
                dict_results[key] = dset_temp[...]

            keys_linear_eigenmodes = (
                self.sim.info.solver.classes.State.keys_linear_eigenmodes
            )

            for key in keys_linear_eigenmodes:
                dset_temp = file[key[:-3] + "array_ik"]
                A = dset_temp[...]
                dict_results["sig_" + key] = np.ascontiguousarray(A.transpose())
        return dict_results

    def plot(self):
        dict_results = self.load()

        t = dict_results["times"]

        nb_shells = dict_results["nb_shells"]
        nb_k_per_shell = dict_results["nb_k_per_shell"]

        sig_q_fft = dict_results["sig_q_fft"]
        sig_a_fft = dict_results["sig_a_fft"]
        sig_d_fft = dict_results["sig_d_fft"]

        kh_shell = dict_results["kh_shell"]
        omega_shell = dict_results["omega_shell"]
        period_shell = 2 * np.pi / omega_shell

        for ish in range(nb_shells):

            fig, ax1 = self.output.figure_axe()
            ax1.set_xlabel("$t/T$")
            ax1.set_ylabel("signals (s$^{-1}$)")
            ax1.set_title(
                "signals eigenmodes, ikh = {:.2f}\n".format(
                    (kh_shell[ish] / self.sim.oper.deltak)
                )
                + self.output.summary_simul
            )

            coef_norm_a = self.c2 / omega_shell[ish]

            T = period_shell[ish]

            for ikps in range(nb_k_per_shell):
                isig = ish * nb_k_per_shell + ikps

                ax1.plot(t / T, sig_q_fft[isig].real, "k", linewidth=1)
                ax1.plot(
                    t / T, coef_norm_a * sig_a_fft[isig].real, "c", linewidth=1
                )
                ax1.plot(t / T, sig_d_fft[isig].real, "y", linewidth=1)

        fig, ax1 = self.output.figure_axe()
        ax1.set_xlabel(r"$\omega$")
        ax1.set_ylabel("kh_shell")
        ax1.loglog(kh_shell, omega_shell, "o", linewidth=2)

    def time_spectrum(self, sig_long):

        Nt = sig_long.size
        stepit0 = int(np.fix(self.nt / 2.0))

        nb_spectra = 0
        it0 = 0
        spect = np.zeros([self.nt // 2 + 1])
        while it0 + self.nt < Nt:
            nb_spectra += 1
            sig = sig_long[it0 : it0 + self.nt]
            spect_raw = (
                abs(self.opfft1d.fft(self.hann * sig)) ** 2 / 2 / self.deltaomega
            )
            spect += spect_raw[: self.nt // 2 + 1]
            if self.nt % 2 == 0:
                spect[1 : self.nt // 2] += spect_raw[
                    self.nt - 1 : self.nt // 2 : -1
                ]
            else:
                spect[1 : self.nt // 2 + 1] += spect_raw[
                    self.nt - 1 : self.nt // 2 : -1
                ]
            it0 += stepit0

        return spect / nb_spectra

    def compute_spectra(self):
        dict_results = self.load()

        t = dict_results["times"]
        Nt = t.size
        nt = 2 ** int(np.fix(np.log2(Nt / 10)))
        if nt < 2:
            nt = 2
        if nt % 2 == 1:
            nt -= 1
        self.nt = nt

        if not hasattr(self, "opfft1d"):
            self.opfft1d = easypyfft.FFTW1D(nt)

        T = t[nt - 1] - t[0]
        # deltat = T/nt
        self.deltaomega = 2 * np.pi / T
        # self.omega = self.deltaomega*np.concatenate(
        #     (np.arange(nt/2+1), np.arange(-nt/2+1, 0)))

        self.omega = self.deltaomega * np.arange(nt // 2 + 1)

        self.hann = np.hanning(nt)

        nb_shells = dict_results["nb_shells"]
        nb_k_per_shell = dict_results["nb_k_per_shell"]
        # nb_k_tot = dict_results['nb_k_tot']

        sig_q_fft = dict_results["sig_q_fft"]
        sig_a_fft = dict_results["sig_a_fft"]
        sig_d_fft = dict_results["sig_d_fft"]

        # kh_shell = dict_results['kh_shell']
        omega_shell = dict_results["omega_shell"]
        # period_shell = 2*np.pi/omega_shell

        time_spectra_q = np.zeros([nb_shells, nt // 2 + 1])
        time_spectra_a = np.zeros([nb_shells, nt // 2 + 1])
        time_spectra_d = np.zeros([nb_shells, nt // 2 + 1])

        for ish in range(nb_shells):
            coef_norm_a = self.c2 / omega_shell[ish]
            for ikps in range(nb_k_per_shell):
                isig = ish * nb_k_per_shell + ikps
                sig_a_fft[isig] *= coef_norm_a
                time_spectra_q[ish] += self.time_spectrum(sig_q_fft[isig])
                time_spectra_a[ish] += self.time_spectrum(sig_a_fft[isig])
                time_spectra_d[ish] += self.time_spectrum(sig_d_fft[isig])

        time_spectra_q /= nb_k_per_shell
        time_spectra_a /= nb_k_per_shell
        time_spectra_d /= nb_k_per_shell

        dict_spectra = {
            "omega": self.omega,
            "time_spectra_q": time_spectra_q,
            "time_spectra_a": time_spectra_a,
            "time_spectra_d": time_spectra_d,
        }
        return dict_spectra, dict_results

    def plot_spectra(self):
        dict_spectra, dict_results = self.compute_spectra()

        omega = dict_spectra["omega"]
        time_spectra_q = dict_spectra["time_spectra_q"]
        time_spectra_a = dict_spectra["time_spectra_a"]
        time_spectra_d = dict_spectra["time_spectra_d"]
        omega_shell = dict_results["omega_shell"]

        fig, ax1 = self.output.figure_axe()
        ax1.set_xlabel(r"r$\omega/\omega_{lin}$")
        ax1.set_ylabel(r"r$E(\omega)$)")
        ax1.set_title("time spectra\n" + self.output.summary_simul)

        nb_shells = dict_results["nb_shells"]
        for ish in range(nb_shells):
            ax1.loglog(
                omega / omega_shell[ish], time_spectra_q[ish], "k", linewidth=1
            )
            ax1.loglog(
                omega / omega_shell[ish], time_spectra_a[ish], "b", linewidth=1
            )
            ax1.loglog(
                omega / omega_shell[ish], time_spectra_d[ish], "r", linewidth=1
            )

    def _close_file(self):
        try:
            self.file.close()
        except AttributeError:
            pass


# if __name__ == '__main__':
#     pass



"""Load and plot benchmarks (:mod:`fluidsim.util.console.bench_analysis`)
=========================================================================

"""
import os
from glob import glob
import json
import sys

import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
from matplotlib.ticker import MaxNLocator

from .bench import path_results, parse_args_dim, init_parser_base, ConsoleError

description = "Plot results of benchmarks"


def load_bench(path_dir, solver, hostname="any"):
    """Load benchmarks results saved as JSON files."""

    if solver.startswith("fluidsim"):
        solver = solver.split(".", 1)[1]

    dicts = []
    for path in glob(path_dir + f"/result_bench_{solver}*.json"):
        with open(path) as file:
            d = json.load(file)

        if hostname != "any" and not d["hostname"].startswith(hostname):
            continue

        dicts.append(d)

    df = pd.DataFrame(dicts)
    df = df[df.columns.difference(["pid", "time_as_str"])]
    return df


def filter_by_shape(df, n0, n1, n2=None):
    """Filters all results with the same `n0` and same `n1`."""

    if n2 is None:
        df = df[(df.n0 == n0) & (df.n1 == n1)]
        return df[df.columns.difference(["n0", "n1"])]

    else:
        df = df[(df.n0 == n0) & (df.n1 == n1) & (df.n2 == n2)]
        return df[df.columns.difference(["n0", "n1", "n2"])]


def filter_by_shapeloc(df, n0_loc, n1_loc, n2_loc=None):
    """Filters all results with the same `n0_loc * n1_loc`.

    This implies shapeK_loc has same no. of points. This is a weak check and
    the shapeK_loc may not have the same shape. Make sure all shapes are
    estimated and tested apriori using:

    >>> fluidsim bench -e
    >>> mpirun -np {nb_proc} fluidsim bench -c

    """
    if n2_loc is None:
        df = df[(df.n0_loc * df.n1_loc == n0_loc * n1_loc)]
        return df[df.columns.difference(["n0_loc", "n1_loc"])]

    else:
        df = df[(df.n0_loc * df.n1_loc * df.n2_loc == n0_loc * n1_loc * n2_loc)]
        return df[df.columns.difference(["n0_loc", "n1_loc", "n2_loc"])]


def exit_if_empty(df, input_params):
    """Check if the dataframe is empty."""
    if df.empty:
        print("No benchmarks corresponding to the input parameters:")
        for k, v in input_params.items():
            print(k, "=", repr(v))
        sys.exit()


def plot_scaling(
    path_dir,
    solver,
    hostname,
    n0,
    n1,
    n2=None,
    show=True,
    type_time="usr",
    type_plot="strong",
    fig=None,
    ax0=None,
    ax1=None,
    name_dir=None,
    once=False,
    t_min_cum=1e10,
):
    """Plot speedup vs number of processes from benchmark results."""

    input_params = dict(
        path_dir=path_dir, solver=solver, hostname=hostname, n0=n0, n1=n1
    )

    if name_dir is None:
        name_dir = (
            os.path.basename(os.path.abspath(path_dir)).replace("_", " ").upper()
        )

    df = load_bench(path_dir, solver, hostname)
    exit_if_empty(df, input_params)
    df.t_elapsed_sys /= df.nb_iter
    df.t_elapsed_usr /= df.nb_iter

    if type_plot == "strong":
        df_filter = filter_by_shape(df, n0, n1, n2)
    elif type_plot == "weak":
        df_filter = filter_by_shapeloc(df, n0, n1, n2)
    else:
        raise ConsoleError("Unknown plot type.")

    def group_df(df):
        """Group and take median dataframe results with same number of processes."""
        # for "scaling" (mpi)
        df = df[df.nb_proc > 1]
        exit_if_empty(df, input_params)
        nb_proc_min = df.nb_proc.min()
        df_grouped = df.groupby(["type_fft", "nb_proc"]).quantile(q=0.2)
        if show:
            print(df)
        return df_grouped, nb_proc_min

    keys_filter = [
        k for k in df_filter.columns if k.startswith("t_elapsed_" + type_time)
    ]

    df_filter, nb_proc_min_filter = group_df(
        df_filter[keys_filter + ["type_fft", "nb_proc"]]
    )
    df_filter = df_filter[keys_filter]
    df_filter_nb_proc_min = df_filter.xs(nb_proc_min_filter, level=1)

    def get_min(df):
        """Get minumum time from the whole dataframe"""
        m = df.values
        i0, i1 = np.unravel_index(np.argmin(m), m.shape)
        mymin = m[i0, i1]
        ind = df.index[i0]
        key = df.columns[i1]
        return mymin, ind, key

    t_min_filter, name_min_filter, key_min_filter = get_min(df_filter_nb_proc_min)

    print(
        "{}: Best for 2 procs t={} for {}, {}".format(
            path_dir, t_min_filter, name_min_filter, key_min_filter
        )
    )

    t_min_filter = min(t_min_filter, t_min_cum)
    # Finally, start preparing figure and plot
    if fig is None or ax0 is None or ax1 is None:
        fig, axes = plt.subplots(1, 2)
        ax0, ax1 = axes.ravel()

    ax0.set_ylabel(f"speedup ({type_plot} scaling)")
    ax1.set_ylabel(f"efficiency % ({type_plot} scaling)")

    def plot_once(ax, x, y, label, linestyle="-k"):
        """Avoid plotting the same label again."""
        plotted = any([lines.get_label() == label for lines in ax.get_lines()])
        if not (plotted and once):
            ax.plot(x, y, linestyle, label=label)

    def add_hline(ax, series=None, y=None):
        """Add a horizontal line to the axis."""
        if y is None:
            y = series.iloc[0]

        xmin = series.index.min()
        xmax = max(series.index.max(), ax.get_xlim()[1])
        # plot_once(ax, [xmin, xmax], [y, y], 'linear')
        ax.plot([xmin, xmax], [y, y], "k-")

    def set_label_scale(ax, yscale="log"):
        """Set log scale, legend and label of the axis."""
        ax.set_xscale("log")
        ax.set_yscale(yscale)
        ax.set_xlabel("number of processes")

    # Plot speedup
    print("Grouped names:")
    for name in df_filter.index.levels[0]:
        tmp = df_filter.loc[name]
        print(name)

        for k in keys_filter:
            speedup = t_min_filter / tmp[k] * nb_proc_min_filter
            ax0.plot(
                speedup.index,
                speedup.values,
                "x-",
                label="{}, {}".format(name.replace("fluidfft.", ""), name_dir),
            )

    ax0.xaxis.set_major_locator(MaxNLocator(integer=True))
    if type_plot == "strong":
        theoretical = [speedup.index.min(), speedup.index.max()]
        # plot_once(ax0, theoretical, theoretical, 'linear')
        ax0.plot(theoretical, theoretical, "k-")
    else:
        add_hline(ax0, speedup, 2)

    set_label_scale(ax0)
    ax0.legend()
    ax0.set_title(
        f"Best for {nb_proc_min_filter} processes: "
        f"{name_min_filter}, {key_min_filter}={t_min_filter * 1000:.2f} ms"
    )

    # Plot efficiency
    for name in df_filter.index.levels[0]:
        tmp = df_filter.loc[name]
        for k in keys_filter:
            speedup = t_min_filter / tmp[k] * nb_proc_min_filter
            if type_plot == "strong":
                efficiency = speedup / speedup.index * 100
            else:
                # efficiency = speedup / speedup.iloc[0] * 100
                efficiency = speedup / 2 * 100
            ax1.plot(
                efficiency.index,
                efficiency.values,
                "x-",
                label=f"{name}, {name_dir}",
            )

    add_hline(ax1, efficiency, 100)

    set_label_scale(ax1, "linear")

    if show:
        plt.show()
    return fig, t_min_filter


def init_parser(parser):
    """Initialize argument parser for `fluidsim bench-analysis`."""

    init_parser_base(parser)
    parser.add_argument(
        "-i",
        "--input-dir",
        default=path_results,
        help="plot results from a single directory",
    )
    parser.add_argument(
        "--input-dirs",
        nargs="+",
        default=[],
        help="plot results from mulitiple input directories",
    )
    parser.add_argument(
        "-p",
        "--type-plot",
        default="strong",
        help="load and plot data for strong or weak scaling analysis",
    )
    parser.add_argument("--save", action="store_true")
    parser.add_argument("--hostname", default="any")


def run(args):
    """Run `fluidsim bench-analysis` command."""
    args = parse_args_dim(args)
    if len(args.input_dirs) == 0:
        args.input_dirs = [args.input_dir]

    fig = plt.figure(figsize=[12, 5])
    ax0 = plt.subplot(121)
    ax1 = plt.subplot(122)
    t_min = 1e10
    for in_dir in args.input_dirs:
        fig, t_min = plot_scaling(
            in_dir,
            args.solver,
            args.hostname,
            args.n0,
            args.n1,
            args.n2,
            show=True,
            type_plot=args.type_plot,
            fig=fig,
            ax0=ax0,
            ax1=ax1,
            t_min_cum=t_min,
        )

    if args.save:
        figname = "fig_bench_" + os.path.basename(args.input_dir) + ".png"
        fig.savefig(figname)
    else:
        plt.show()



from .solver import Simul

from fluidsim.solvers.ad1d.test_solver import TestSolverAD1D
from fluidsim.util.testing import skip_if_no_fluidfft


@skip_if_no_fluidfft
class TestSolverAD1DPseudoSpectral(TestSolverAD1D):
    Simul = Simul


del TestSolverAD1D



import numpy as np

from fluidsim.base.time_stepping.pseudo_spect import TimeSteppingPseudoSpectral
from fluidsim.operators.operators3d import dealiasing_variable


class TimeSteppingPseudoSpectralNS3D(TimeSteppingPseudoSpectral):
    def one_time_step_computation(self):
        """One time step"""
        state_spect = self.sim.state.state_spect
        # WARNING: if the function _time_step_RK comes from an extension, its
        # execution time seems to be attributed to the function
        # one_time_step_computation by cProfile
        self._time_step_RK()
        self.sim.oper.dealiasing(state_spect)
        vx_fft = state_spect.get_var("vx_fft")
        vy_fft = state_spect.get_var("vy_fft")
        vz_fft = state_spect.get_var("vz_fft")
        self.sim.oper.project_perpk3d(vx_fft, vy_fft, vz_fft)
        if self.sim.no_vz_kz0:
            dealiasing_variable(vz_fft, self.sim.where_kz_0)
            if "b_fft" in state_spect.keys:
                dealiasing_variable(
                    state_spect.get_var("b_fft"), self.sim.where_kz_0
                )

        self.sim.state.statephys_from_statespect()
        # np.isnan(np.sum seems to be really fast
        if np.isnan(np.sum(state_spect[0])):
            raise ValueError(f"nan at it = {self.it}, t = {self.t:.4f}")



"""
SW1L forcing (:mod:`fluidsim.solvers.sw1l.forcing`)
===================================================

"""

import numpy as np

from fluiddyn.util import mpi

from fluidsim.base.forcing import ForcingBasePseudoSpectral

from fluidsim.base.forcing.specific import (
    InScriptForcingPseudoSpectral,
    Proportional as ProportionalBase,
    TimeCorrelatedRandomPseudoSpectral as TCRandomPS,
    RandomSimplePseudoSpectral,
)


class ForcingSW1L(ForcingBasePseudoSpectral):
    """Forcing class for the sw1l solver.

    .. inheritance-diagram:: ForcingSW1L

    """

    @staticmethod
    def _complete_info_solver(info_solver):
        """Complete the ParamContainer info_solver.

        This is a static method!
        """
        classes = [
            TCRandomPS,
            RandomSimplePseudoSpectral,
            Waves,
            WavesVortices,
            Potential,
        ]
        ForcingBasePseudoSpectral._complete_info_solver(info_solver, classes)


class TimeCorrelatedRandomPseudoSpectral(TCRandomPS):
    """Forces the geostrophic variable (by default) while maintaining a
    correlation for a certain time interval.

    """

    _key_forced_default = "q_fft"


class Proportional(ProportionalBase):
    _key_forced_default = "q_fft"


class Waves(RandomSimplePseudoSpectral):
    """Forces the ageostrophic variable and normalizes the forcing power based
    on the K.E and A.P.E thus generated. The forcing is white noise in time.

    """

    tag = "waves"
    _key_forced_default = "a_fft"

    def normalize_forcingc_2nd_degree_eq(self, Fa_fft, a_fft):
        """Normalize the forcing Fa_fft such that the forcing rate of
        quadratic energy is equal to self.forcing_rate."""
        oper_c = self.oper_coarse
        params = self.params
        deltat = self.sim.time_stepping.deltat

        Fux_fft, Fuy_fft, Feta_fft = oper_c.uxuyetafft_from_afft(Fa_fft)
        ux_fft, uy_fft, eta_fft = oper_c.uxuyetafft_from_afft(a_fft)

        ax = deltat / 2 * oper_c.sum_wavenumbers(abs(Fux_fft) ** 2)
        ay = deltat / 2 * oper_c.sum_wavenumbers(abs(Fuy_fft) ** 2)
        aA = params.c2 * deltat / 2 * oper_c.sum_wavenumbers(abs(Feta_fft) ** 2)
        a = ax + ay + aA

        bx = oper_c.sum_wavenumbers((ux_fft.conj() * Fux_fft).real)
        by = oper_c.sum_wavenumbers((uy_fft.conj() * Fuy_fft).real)
        bA = params.c2 * oper_c.sum_wavenumbers((eta_fft.conj() * Feta_fft).real)
        b = bx + by + bA

        c = -self.forcing_rate

        alpha = self.coef_normalization_from_abc(a, b, c)
        Fa_fft[:] = alpha * Fa_fft

        return Fa_fft


class WavesVortices(Waves):
    """Forces both the geostrophic and ageostrophic variable equally and
    normalizes the forcing rate based on the K.E and A.P.E thus generated.

    """

    tag = "waves_vortices"
    _key_forced_default = ("q_fft", "a_fft")

    def __init__(self, sim):
        super().__init__(sim)
        params = sim.params.forcing
        self.forcing_rate = 0.5 * params.forcing_rate

    def compute(self):
        if not isinstance(self.key_forced, (tuple, list, np.ndarray)):
            raise ValueError(
                "Expected array-like value for params.forcing.key_forced"
                + " = {} : {}".format(self.key_forced, type(self.key_forced))
            )

        v_fft = dict()
        Fv_fft = dict()

        for key in self.key_forced:
            try:
                v_fft[key] = self.sim.state.state_spect.get_var(key)
            except ValueError:
                v_fft[key] = self.sim.state.get_var(key)

            v_fft[key] = self.oper.coarse_seq_from_fft_loc(
                v_fft[key], self.shapeK_loc_coarse
            )

            if mpi.rank == 0:
                Fv_fft["F" + key] = self.forcingc_raw_each_time(key)

        if mpi.rank == 0:
            kwargs = v_fft
            kwargs.update(Fv_fft)
            Fv_fft = self.normalize_forcingc(**kwargs)
            self.fstate_coarse.init_statespect_from(**Fv_fft)

        self.put_forcingc_in_forcing()

    def normalize_forcingc(self, Fq_fft, Fa_fft, q_fft, a_fft):
        """Normalize the forcing Fa_fft such as the forcing rate of
        quadratic energy is equal to self.forcing_rate."""

        Fa_fft = self.normalize_forcingc_2nd_degree_eq(Fa_fft, a_fft)

        oper_c = self.oper_coarse
        params = self.params
        deltat = self.sim.time_stepping.deltat

        Fux_fft, Fuy_fft, Feta_fft = oper_c.uxuyetafft_from_qfft(Fq_fft)
        ux_fft, uy_fft, eta_fft = oper_c.uxuyetafft_from_qfft(q_fft)

        ax = deltat / 2 * oper_c.sum_wavenumbers(abs(Fux_fft) ** 2)
        ay = deltat / 2 * oper_c.sum_wavenumbers(abs(Fuy_fft) ** 2)
        aA = params.c2 * deltat / 2 * oper_c.sum_wavenumbers(abs(Feta_fft) ** 2)
        a = ax + ay + aA

        bx = oper_c.sum_wavenumbers((ux_fft.conj() * Fux_fft).real)
        by = oper_c.sum_wavenumbers((uy_fft.conj() * Fuy_fft).real)
        bA = params.c2 * oper_c.sum_wavenumbers((eta_fft.conj() * Feta_fft).real)
        b = bx + by + bA

        c = -self.forcing_rate

        alpha = self.coef_normalization_from_abc(a, b, c)
        Fq_fft[:] = alpha * Fq_fft

        Fv_fft = {"q_fft": Fq_fft, "a_fft": Fa_fft}

        return Fv_fft


class Potential(Waves):
    """Forces only in A.P.E. and normalize for the desired forcing rate."""

    tag = "potential"
    _key_forced_default = "eta_fft"

    def normalize_forcingc_2nd_degree_eq(self, Feta_fft, eta_fft):
        """Normalize the forcing Fa_fft such as the forcing rate of
        quadratic energy is equal to self.forcing_rate."""
        if "eta_fft" not in self.key_forced:
            raise ValueError(
                "Expected 'eta_fft' in params.forcing.key_forced = {}".format(
                    self.key_forced
                )
            )

        oper_c = self.oper_coarse
        params = self.params
        deltat = self.sim.time_stepping.deltat

        a = params.c2 * deltat / 2 * oper_c.sum_wavenumbers(abs(Feta_fft) ** 2)

        b = params.c2 * oper_c.sum_wavenumbers((eta_fft.conj() * Feta_fft).real)

        c = -self.forcing_rate

        alpha = self.coef_normalization_from_abc(a, b, c)
        Feta_fft[:] = alpha * Feta_fft

        return Feta_fft


# class OldStuff(object):

#     def verify_injection_rate_opfft(self, q_fft, Fq_fft, oper):
#         """Verify injection rate."""
#         P_Z_forcing1 = abs(Fq_fft)**2/2*self.sim.time_stepping.deltat
#         P_Z_forcing2 = np.real(Fq_fft.conj()*q_fft)
#         P_Z_forcing1 = oper.sum_wavenumbers(P_Z_forcing1)
#         P_Z_forcing2 = oper.sum_wavenumbers(P_Z_forcing2)
#         if mpi.rank == 0:
#             print('P_Z_f = {0:9.4e} ; P_Z_f2 = {1:9.4e};'.format(
#                 P_Z_forcing1+P_Z_forcing2,
#                 P_Z_forcing2))

#     def verify_injection_rate_from_state(self):
#         """Verify injection rate."""

#         ux_fft = self.sim.state.state_spect.get_var('ux_fft')
#         uy_fft = self.sim.state.state_spect.get_var('uy_fft')
#         eta_fft = self.sim.state.state_spect.get_var('eta_fft')

#         q_fft, div_fft, ageo_fft = \
#             self.oper.qdafft_from_uxuyetafft(ux_fft, uy_fft, eta_fft)

#         Fux_fft = self.forcing_fft.get_var('ux_fft')
#         Fuy_fft = self.forcing_fft.get_var('uy_fft')
#         Feta_fft = self.forcing_fft.get_var('eta_fft')

#         Fq_fft, Fdiv_fft, Fageo_fft = \
#             self.oper.qdafft_from_uxuyetafft(Fux_fft, Fuy_fft, Feta_fft)
#         # print 'Fq_fft', abs(Fq_fft).max()
#         # print 'Fdiv_fft', abs(Fdiv_fft).max()
#         # print 'Fageo_fft', abs(Fageo_fft).max()

#         self.verify_injection_rate_opfft(q_fft, Fq_fft, self.oper)

#     def modify_Ffft_from_eta(self, F_fft, eta_fft):
#         """Put to zero the forcing for the too large modes."""
#         for ik in self.ind_forcing:
#             if abs(eta_fft.flat[ik]) > self.eta_cond:
#                 F_fft.flat[ik] = 0.



"""Operators plate2d model (:mod:`fluidsim.operators.operators_plate2d`)
========================================================================

Provides

.. autoclass:: OperatorsPseudoSpectralPlate2D
   :members:
   :private-members:

"""

import numpy as np

from transonic import jit, Array

from ...operators.operators2d import OperatorsPseudoSpectral2D

AC2 = Array[np.complex128, "2d"]
AF2 = Array[np.float64, "2d"]


@jit
def monge_ampere_step0(a_fft: AC2, b_fft: AC2, KX2: AF2, KY2: AF2, KXKZ: AF2):
    pxx_a_fft = -a_fft * KX2
    pyy_a_fft = -a_fft * KY2
    pxy_a_fft = -a_fft * KXKZ
    pxx_b_fft = -b_fft * KX2
    pyy_b_fft = -b_fft * KY2
    pxy_b_fft = -b_fft * KXKZ
    return pxx_a_fft, pyy_a_fft, pxy_a_fft, pxx_b_fft, pyy_b_fft, pxy_b_fft


@jit
def monge_ampere_step1(
    pxx_a: AF2, pyy_a: AF2, pxy_a: AF2, pxx_b: AF2, pyy_b: AF2, pxy_b: AF2
):
    return pxx_a * pyy_b + pyy_a * pxx_b - 2 * pxy_a * pxy_b


class OperatorsPseudoSpectralPlate2D(OperatorsPseudoSpectral2D):
    """Operators for the plate2d model."""

    def __init__(self, params):
        super().__init__(params)

        self.KX2 = self.KX ** 2
        self.KY2 = self.KY ** 2
        self.KXKY = self.KX * self.KY

        self.tmp_pxx_a = np.empty_like(self.XX)
        self.tmp_pyy_a = np.empty_like(self.XX)
        self.tmp_pxy_a = np.empty_like(self.XX)

        self.tmp_pxx_b = np.empty_like(self.XX)
        self.tmp_pyy_b = np.empty_like(self.XX)
        self.tmp_pxy_b = np.empty_like(self.XX)

    def monge_ampere_from_fft(self, a_fft, b_fft):
        """Compute the Monge-Ampere operator"""

        (
            pxx_a_fft,
            pyy_a_fft,
            pxy_a_fft,
            pxx_b_fft,
            pyy_b_fft,
            pxy_b_fft,
        ) = monge_ampere_step0(a_fft, b_fft, self.KX2, self.KY2, self.KXKY)

        self.ifft_as_arg(pxx_a_fft, self.tmp_pxx_a)
        self.ifft_as_arg(pyy_a_fft, self.tmp_pyy_a)
        self.ifft_as_arg(pxy_a_fft, self.tmp_pxy_a)
        self.ifft_as_arg(pxx_b_fft, self.tmp_pxx_b)
        self.ifft_as_arg(pyy_b_fft, self.tmp_pyy_b)
        self.ifft_as_arg(pxy_b_fft, self.tmp_pxy_b)

        return monge_ampere_step1(
            self.tmp_pxx_a,
            self.tmp_pyy_a,
            self.tmp_pxy_a,
            self.tmp_pxx_b,
            self.tmp_pyy_b,
            self.tmp_pxy_b,
        )


# def monge_ampere_from_fft_numpy(a_fft, b_fft, KX, KY, ifft2):

#     pxx_a = -ifft2(a_fft * KX ** 2)
#     pyy_a = -ifft2(a_fft * KY ** 2)
#     pxy_a = -ifft2(a_fft * KX * KY)

#     pxx_b = -ifft2(b_fft * KX ** 2)
#     pyy_b = -ifft2(b_fft * KY ** 2)
#     pxy_b = -ifft2(b_fft * KX * KY)

#     return pxx_a * pyy_b + pyy_a * pxx_b - 2 * pxy_a * pxy_b



"""Operators for Dedalus (:mod:`fluidsim.base.dedalus.operators`)
===================================================================

Provides:

.. autoclass:: OperatorsDedalus2D
   :members:
   :private-members:

"""

import numpy as np

from dedalus import public as dedalus


class OperatorsDedalus2D:
    """2D operators."""

    @staticmethod
    def _complete_params_with_default(params):
        """This static method is used to complete the *params* container."""

        attribs = {"nx": 128, "nz": 64, "Lx": 4, "Lz": 1}
        params._set_child("oper", attribs=attribs)

    def create_domain(self, params):
        de = dedalus
        par_oper = params.oper

        # Create bases and domain
        x_basis = de.Fourier(
            "x", par_oper.nx, interval=(0, par_oper.Lx), dealias=3 / 2
        )
        z_basis = de.Chebyshev(
            "z",
            par_oper.nz,
            interval=(-par_oper.Lz / 2, par_oper.Lz / 2),
            dealias=3 / 2,
        )
        return de.Domain([x_basis, z_basis], grid_dtype=np.float64)

    def __init__(self, params=None, SEQUENTIAL=None):

        self.domain = self.create_domain(params)

        self.axes = ("z", "x")
        self.Lx = params.oper.Lx
        self.Ly = params.oper.Lz

        self.params = params

        self.x_seq, self.y_seq = self.domain.grids(scales=1.0)
        self.x_seq = self.x_seq.flatten()
        self.y_seq = self.y_seq.flatten()
        self.X, self.Y = np.meshgrid(self.x_seq, self.y_seq)

        nx = self.nx_seq = len(self.x_seq)
        ny = self.ny_seq = len(self.y_seq)

        self.shapeX_seq = self.shapeX_loc = (ny, nx)

    def get_grid1d_seq(self, axe="x"):

        if axe not in ("x", "y"):
            raise ValueError

        return getattr(self, axe + "_seq")

    def produce_str_describing_oper(self):
        """Produce a string describing the operator."""

        return ""

    def produce_long_str_describing_oper(self):
        """Produce a string describing the operator."""

        return "2d Dedalus simulation\n"



"""Time stepping (:mod:`fluidsim.base.time_stepping.pseudo_spect`)
========================================================================

Provides:

.. autoclass:: TimeSteppingPseudoSpectral
   :members:
   :private-members:

.. todo::

  It would be interesting to also implement the Adams-Bashforth (leapfrog)
  scheme with phase-shifting. It is very close to
  :func:`_time_step_RK2_phaseshift` with 2 evaluations of the non-linear terms
  per time step (but with 2 symmetrical and equivalent steps).

.. note::

  For a theoretical presentation of phase-shifting, see the book Numerical
  Experiments in Homogeneous Turbulence (Robert S. Rogallo,
  https://ntrs.nasa.gov/archive/nasa/casi.ntrs.nasa.gov/19810022965.pdf).

"""
from random import randint

import numpy as np

from transonic import Transonic, Type, NDim, Array, boost, Union

from .base import TimeSteppingBase

ts = Transonic()

N = NDim(2, 3, 4)
A = Array[np.complex128, N, "C"]
Am1 = Array[np.complex128, N - 1, "C"]

N123 = NDim(1, 2, 3)
A123c = Array[np.complex128, N123, "C"]
A123f = Array[np.float64, N123, "C"]

T = Type(np.float64, np.complex128)
A1 = Array[T, N, "C"]
A2 = Array[T, N - 1, "C"]
ArrayDiss = Union[A1, A2]

uniform = np.random.default_rng().uniform


@boost
def step_Euler(
    state_spect: A, dt: float, tendencies: A, diss: ArrayDiss, output: A
):
    output[:] = (state_spect + dt * tendencies) * diss
    return output


@boost
def step_Euler_inplace(state_spect: A, dt: float, tendencies: A, diss: ArrayDiss):
    step_Euler(state_spect, dt, tendencies, diss, state_spect)


@boost
def step_like_RK2(
    state_spect: A, dt: float, tendencies: A, diss: ArrayDiss, diss2: ArrayDiss
):
    state_spect[:] = state_spect * diss + dt * diss2 * tendencies


@boost
def mean_with_phaseshift(
    tendencies_0: A, tendencies_1_shift: A, phaseshift: Am1, output: A
):
    output[:] = 0.5 * (tendencies_0 + tendencies_1_shift / phaseshift)
    return output


@boost
def mul(phaseshift: Am1, state_spect: A, output: A):
    output[:] = phaseshift * state_spect
    return output


@boost
def div_inplace(arr: A, phaseshift: Am1):
    arr /= phaseshift
    return arr


@boost
def compute_phaseshift_terms(
    phase_alpha: A123f,
    phase_beta: A123f,
    phaseshift_alpha: A123c,
    phaseshift_beta: A123c,
):
    phaseshift_alpha[:] = np.exp(1j * phase_alpha)
    phaseshift_beta[:] = np.exp(1j * phase_beta)
    return phaseshift_alpha, phaseshift_beta


class ExactLinearCoefs:
    """Handle the computation of the exact coefficient for the RK4."""

    def __init__(self, time_stepping):
        self.time_stepping = time_stepping
        sim = time_stepping.sim
        self.shapeK_loc = sim.oper.shapeK_loc
        self.freq_lin = time_stepping.freq_lin

        self.exact = np.empty_like(self.freq_lin)
        self.exact2 = np.empty_like(self.freq_lin)

        if sim.params.time_stepping.USE_CFL:
            self.get_updated_coefs = self.get_updated_coefs_CLF
            self.dt_old = 0.0
        else:
            self.compute(time_stepping.deltat)
            self.get_updated_coefs = self.get_coefs

    def compute(self, dt):
        """Compute the exact coefficients."""
        f_lin = self.freq_lin
        exact = self.exact
        exact2 = self.exact2

        if ts.is_transpiled:
            ts.use_block("exact_lin_compute")
        else:
            # transonic block (
            #     A1 f_lin, exact, exact2;
            #     float dt
            # )
            # transonic block (
            #     A2 f_lin, exact, exact2;
            #     float dt
            # )
            exact[:] = np.exp(-dt * f_lin)
            exact2[:] = np.exp(-dt / 2 * f_lin)
        self.dt_old = dt

    def get_updated_coefs_CLF(self):
        """Get the exact coefficient updated if needed."""
        dt = self.time_stepping.deltat
        if self.dt_old != dt:
            self.compute(dt)
        return self.exact, self.exact2

    def get_coefs(self):
        """Get the exact coefficients as stored."""
        return self.exact, self.exact2


class TimeSteppingPseudoSpectral(TimeSteppingBase):
    """Time stepping class for pseudo-spectral solvers."""

    @staticmethod
    def _complete_params_with_default(params):
        """This static method is used to complete the *params* container."""
        TimeSteppingBase._complete_params_with_default(params)
        params.time_stepping.USE_CFL = True

        params.time_stepping._set_child(
            "phaseshift_random",
            attribs=dict(nb_pairs=1, nb_steps_compute_new_pair=None),
        )

    def __init__(self, sim):
        super().__init__(sim)
        self.init_from_params()

    def init_from_params(self):
        self._init_freq_lin()
        self._init_compute_time_step()
        self._init_exact_linear_coef()
        self._init_time_scheme()

    def _init_freq_lin(self):
        f_d, f_d_hypo = self.sim.compute_freq_diss()
        freq_dissip = f_d + f_d_hypo

        if hasattr(self.sim, "compute_freq_complex"):
            freq_complex = self._compute_freq_complex()
            self.freq_lin = freq_dissip + freq_complex
            freq_max = freq_complex.imag.max()
            self.deltat_max = 0.78 * np.pi / freq_max
        else:
            self.freq_lin = freq_dissip

    def _init_time_scheme(self):

        type_time_scheme = self.params.time_stepping.type_time_scheme

        if type_time_scheme.startswith("RK"):
            self._state_spect_tmp = np.empty_like(self.sim.state.state_spect)

        if type_time_scheme.endswith("_random"):
            self._init_phaseshift_random()
            if not hasattr(self.sim.oper, "get_phases_random"):
                raise NotImplementedError

        if type_time_scheme == "Euler":
            time_step_RK = self._time_step_Euler
        elif type_time_scheme == "Euler_phaseshift":
            time_step_RK = self._time_step_Euler_phaseshift
        elif type_time_scheme == "Euler_phaseshift_random":
            time_step_RK = self._time_step_Euler_phaseshift_random
        elif type_time_scheme == "RK2":
            time_step_RK = self._time_step_RK2
        elif type_time_scheme == "RK2_trapezoid":
            time_step_RK = self._time_step_RK2_trapezoid
        elif type_time_scheme == "RK2_phaseshift":
            time_step_RK = self._time_step_RK2_phaseshift
        elif type_time_scheme == "RK2_phaseshift_random":
            time_step_RK = self._time_step_RK2_phaseshift_random
        elif type_time_scheme == "RK2_phaseshift_exact":
            time_step_RK = self._time_step_RK2_phaseshift_exact
        elif type_time_scheme == "RK4":
            self._state_spect_tmp1 = np.empty_like(self.sim.state.state_spect)
            time_step_RK = self._time_step_RK4
        else:
            raise ValueError(f'Problem name time_scheme ("{type_time_scheme}")')

        self._time_step_RK = time_step_RK

    def _compute_freq_complex(self):
        state_spect = self.sim.state.state_spect
        freq_complex = np.empty_like(state_spect)
        for ik, key in enumerate(state_spect.keys):
            freq_complex[ik] = self.sim.compute_freq_complex(key)
        return freq_complex

    def _init_exact_linear_coef(self):
        self.exact_linear_coefs = ExactLinearCoefs(self)

    def one_time_step_computation(self):
        """One time step."""
        self._time_step_RK()
        self.sim.oper.dealiasing(self.sim.state.state_spect)
        self.sim.state.statephys_from_statespect()
        # np.isnan(np.sum seems to be really fast
        if np.isnan(np.sum(self.sim.state.state_spect[0])):
            raise ValueError(f"nan at it = {self.it}, t = {self.t:.4f}")

    def _time_step_Euler(self):
        r"""Forward Euler method.

        Notes
        -----

        .. |p| mathmacro:: \partial
        .. |dt| mathmacro:: \mathop{dt}
        .. |dx| mathmacro:: \mathop{dx}

        We consider an equation of the form

        .. math:: \p_t S = \sigma S + N(S),

        The forward Euler method computes an approximation of the
        solution after a time increment :math:`\dt`. We denote the
        initial time :math:`t = 0`.

        Euler approximation :

          .. math:: \p_t \log S = \sigma + \frac{N_0}{S_0},

        Integrating from :math:`t` to :math:`t+\dt`, it gives:

          .. math:: S_{\dt} = (S_0 + N_0 \dt) e^{\sigma \dt}.

        """
        dt = self.deltat
        diss = self.exact_linear_coefs.get_updated_coefs()[0]

        compute_tendencies = self.sim.tendencies_nonlin
        state_spect = self.sim.state.state_spect

        tendencies_0 = compute_tendencies()
        step_Euler_inplace(state_spect, dt, tendencies_0, diss)

    def _get_phaseshift(self):
        """Compute the phase shift term."""
        if hasattr(self, "_phaseshift") and self._phaseshift is not None:
            return self._phaseshift
        oper = self.sim.oper
        ndim = len(oper.axes)
        if ndim == 1:
            phase = 0.5 * oper.deltax * oper.kx
        elif ndim == 2:
            phase = 0.5 * (oper.deltax * oper.KX + oper.deltay * oper.KY)
        elif ndim == 3:
            phase = 0.5 * (
                oper.deltax * oper.Kx
                + oper.deltay * oper.Ky
                + oper.deltaz * oper.Kz
            )
        else:
            raise NotImplementedError
        self._phaseshift = np.exp(1j * phase)
        return self._phaseshift

    def _init_phaseshift_random(self):

        params_phaseshift = self.params.time_stepping.phaseshift_random
        if params_phaseshift.nb_steps_compute_new_pair is None:
            if params_phaseshift.nb_pairs == 1:
                params_phaseshift.nb_steps_compute_new_pair = 2
            else:
                params_phaseshift.nb_steps_compute_new_pair = (
                    4 * params_phaseshift.nb_pairs
                )

        self._index_phaseshift = 1
        self._previous_index_pair = 0
        self._previous_index_flip = 0
        self._pairs_phaseshift = []
        for _ in range(params_phaseshift.nb_pairs):
            phaseshift_alpha = np.empty(
                self.sim.oper.shapeK_loc, dtype=np.complex128
            )
            phaseshift_beta = np.empty_like(phaseshift_alpha)
            phase_alpha, phase_beta = self.sim.oper.get_phases_random()

            self._pairs_phaseshift.append(
                compute_phaseshift_terms(
                    phase_alpha, phase_beta, phaseshift_alpha, phaseshift_beta
                )
            )

    def _get_phaseshift_random(self):
        """Compute two random phase shift terms."""
        params_phaseshift = self.params.time_stepping.phaseshift_random

        nb_pairs = params_phaseshift.nb_pairs
        nb_steps_compute_new_pair = params_phaseshift.nb_steps_compute_new_pair

        if nb_pairs == 1 and nb_steps_compute_new_pair == 1:
            phaseshift_alpha, phaseshift_beta = self._pairs_phaseshift[0]
        elif nb_pairs == 1 and nb_steps_compute_new_pair == 2:
            pair = self._pairs_phaseshift[0]
            if self._index_phaseshift == 1:
                phaseshift_alpha, phaseshift_beta = pair
            else:
                phaseshift_beta, phaseshift_alpha = pair
        else:
            index_pair = randint(0, nb_pairs - 1)
            pair = self._pairs_phaseshift[index_pair]
            index_flip = randint(0, 1)
            if (
                index_pair == self._previous_index_pair
                and index_flip == self._previous_index_flip
            ):
                index_flip = 0 if index_flip else 1
            self._previous_index_pair = index_pair
            self._previous_index_flip = index_flip
            if index_flip:
                phaseshift_alpha, phaseshift_beta = pair
            else:
                phaseshift_beta, phaseshift_alpha = pair

        if self._index_phaseshift == nb_steps_compute_new_pair:
            self._index_phaseshift = 1
            phase_alpha, phase_beta = self.sim.oper.get_phases_random()
            phaseshift_alpha, phaseshift_beta = self._pairs_phaseshift.pop(0)
            self._pairs_phaseshift.append(
                compute_phaseshift_terms(
                    phase_alpha, phase_beta, phaseshift_alpha, phaseshift_beta
                )
            )
        else:
            self._index_phaseshift += 1

        return phaseshift_alpha, phaseshift_beta

    def _time_step_Euler_phaseshift(self):
        r"""Forward Euler method, dealiasing with phase-shifting.

        Notes
        -----

        We consider an equation of the form

        .. math:: \p_t S = \sigma S + N(S),

        The forward Euler method computes an approximation of the
        solution after a time increment :math:`\dt`. We denote the
        initial time :math:`t = 0`.

        - Euler approximation:

          .. math:: \p_t \log S = \sigma + \frac{N_\mathrm{dealias}}{S_0},

          Integrating from :math:`t` to :math:`t+\dt`, it gives:

          .. math:: S_{\dt} = (S_0 + N_\mathrm{dealias} \dt) e^{\sigma \dt}.

          where the dealiased non-linear term is :math:`N_\mathrm{dealias} =
          (N_0 + \tilde N_0)/2` and the phase-shifted nonlinear term
          :math:`\tilde N_0` is given by

          .. math::
            \tilde N_0 = e^{-\frac{1}{2}k\dx}N\left(e^{\frac{1}{2}k\dx}S_0\right).

        """
        dt = self.deltat
        diss = self.exact_linear_coefs.get_updated_coefs()[0]

        compute_tendencies = self.sim.tendencies_nonlin
        state_spect = self.sim.state.state_spect

        # regular tendencies
        tendencies_0 = compute_tendencies()
        # shifted tendencies
        phaseshift = self._get_phaseshift()
        tendencies_shifted = (
            compute_tendencies(phaseshift * state_spect) / phaseshift
        )
        # dealiased tendencies
        tendencies_dealiased = 0.5 * (tendencies_0 + tendencies_shifted)
        step_Euler_inplace(state_spect, dt, tendencies_dealiased, diss)

    def _time_step_Euler_phaseshift_random(self):
        r"""Forward Euler method, dealiasing with phase-shifting.

        Notes
        -----

        We consider an equation of the form

        .. math:: \p_t S = \sigma S + N(S),

        The forward Euler method computes an approximation of the
        solution after a time increment :math:`\dt`. We denote the
        initial time :math:`t = 0`.

        - Euler approximation:

          .. math:: \p_t \log S = \sigma + \frac{N_\mathrm{dealias}}{S_0},

          Integrating from :math:`t` to :math:`t+\dt`, it gives:

          .. math:: S_{\dt} = (S_0 + N_\mathrm{dealias} \dt) e^{\sigma \dt}.

          where the dealiased non-linear term :math:`N_\mathrm{dealias} =
          (\tilde N_{0\alpha} + \tilde N_{0\beta})/2` is computed as the
          average of two terms shifted with dependant phases
          :math:`\phi_\alpha = \alpha \dx k` and :math:`\phi_\beta = \beta \dx
          k` with :math:`\alpha` taken randomly between -1 and 1 and
          :math:`|\alpha - \beta| = 0.5`.

        """
        dt = self.deltat
        diss = self.exact_linear_coefs.get_updated_coefs()[0]

        compute_tendencies = self.sim.tendencies_nonlin
        state_spect = self.sim.state.state_spect

        # shifted tendencies
        phaseshift_alpha, phaseshift_beta = self._get_phaseshift_random()
        tendencies_alpha = (
            compute_tendencies(phaseshift_alpha * state_spect) / phaseshift_alpha
        )
        tendencies_beta = (
            compute_tendencies(phaseshift_beta * state_spect) / phaseshift_beta
        )
        # dealiased tendencies
        tendencies_dealiased = 0.5 * (tendencies_alpha + tendencies_beta)
        step_Euler_inplace(state_spect, dt, tendencies_dealiased, diss)

    def _time_step_RK2(self):
        r"""Runge-Kutta 2 method.

        Notes
        -----

        We consider an equation of the form

        .. math:: \p_t S = \sigma S + N(S),

        The Runge-Kutta 2 method computes an approximation of the
        solution after a time increment :math:`\dt`. We denote the
        initial time :math:`t = 0`.

        - Approximation 1:

          .. math:: \p_t \log S = \sigma + \frac{N_0}{S_0},

          Integrating from :math:`t` to :math:`t+\dt/2`, it gives:

          .. math:: S_1 = (S_0 + \frac{\dt}{2} N_0) e^{\sigma \frac{\dt}{2}}.

        - Approximation 2:

          .. math::
             \p_t \log S = \sigma + \frac{N_1}{S_1},

          Integrating from :math:`t` to :math:`t+\dt` and retaining
          only the terms in :math:`(N\dt/S)^1` gives:

          .. math::
             S_2 = S_0 e^{\sigma \dt} + \dt N_1 e^{\sigma \frac{\dt}{2}}.

        """
        dt = self.deltat
        diss, diss2 = self.exact_linear_coefs.get_updated_coefs()

        compute_tendencies = self.sim.tendencies_nonlin
        state_spect = self.sim.state.state_spect

        tendencies_0 = compute_tendencies()

        state_spect_12 = self._state_spect_tmp
        step_Euler(
            state_spect, dt / 2, tendencies_0, diss2, output=state_spect_12
        )

        tendencies_12 = compute_tendencies(state_spect_12, old=tendencies_0)
        step_like_RK2(state_spect, dt, tendencies_12, diss, diss2)

    def _time_step_RK2_trapezoid(self):
        r"""Runge-Kutta 2 method with trapezoidal rule (Heun's method).

        Notes
        -----

        We consider an equation of the form

        .. math:: \p_t S = \sigma S + N(S),

        Heun's method computes an approximation of the
        solution after a time increment :math:`\dt`. We denote the
        initial time :math:`t = 0`.

        - Approximation 1:

          .. math:: \p_t \log S = \sigma + \frac{N(S_0)}{S_0},

          Integrating from :math:`t` to :math:`t+\dt`, it gives:

          .. math:: S_1 = (S_0 + N_0 \dt) e^{\sigma \dt}.

        - Approximation 2:

          .. math::
             \p_t \log S = \sigma + \frac{1}{2}\left(
                 \frac{N_0}{S_0} + \frac{N_1}{S_1}\right),

          Integrating from :math:`t` to :math:`t+\dt` and retaining
          only the terms in :math:`(N\dt/S)^1` gives:

          .. math::
             S_2 = S_0 e^{\sigma \dt} + \frac{\dt}{2} (N_0 e^{\sigma \dt} + N_1).

        """
        dt = self.deltat
        diss, diss2 = self.exact_linear_coefs.get_updated_coefs()

        compute_tendencies = self.sim.tendencies_nonlin
        state_spect = self.sim.state.state_spect

        tendencies_0 = compute_tendencies()

        state_spect_1 = self._state_spect_tmp
        step_Euler(state_spect, dt, tendencies_0, diss, output=state_spect_1)

        tendencies_1 = compute_tendencies(state_spect_1)

        state_spect[:] = (
            state_spect + dt / 2 * tendencies_0
        ) * diss + dt / 2 * tendencies_1

    def _time_step_RK2_phaseshift(self):
        r"""Runge-Kutta 2 method with phase-shifting.

        Notes
        -----

        We consider an equation of the form

        .. math:: \p_t S = \sigma S + N(S),

        Heun's method computes an approximation of the
        solution after a time increment :math:`\dt`. We denote the
        initial time :math:`t = 0`.

        - Approximation 1:

          .. math:: \p_t \log S = \sigma + \frac{N_0}{S_0},

          Integrating from :math:`t` to :math:`t+\dt`, it gives:

          .. math:: S_1 = (S_0 + N_0 \dt) e^{\sigma \dt}.

        - Approximation 2:

          .. math::
             \p_t \log S = \sigma + \frac{N_d}{S_0 e^{\sigma \frac{\dt}{2}}},


          where the dealiased non-linear term is :math:`N_\mathrm{dealias} =
          (N_0 + \tilde N_1)/2` and the phase-shifted nonlinear term
          :math:`\tilde N_1` is given by

          .. math::
            \tilde N_1 = e^{-\frac{1}{2}k\dx}N\left(e^{\frac{1}{2}k\dx}S_1\right).

          Integrating from :math:`t` to :math:`t+\dt` and retaining
          only the terms in :math:`(N\dt/S)^1` gives:

          .. math::
             S_2 = S_0 e^{\sigma \dt} + \dt N_d e^{\sigma \frac{\dt}{2}}.

        """
        dt = self.deltat
        diss, diss2 = self.exact_linear_coefs.get_updated_coefs()

        compute_tendencies = self.sim.tendencies_nonlin
        state_spect = self.sim.state.state_spect

        tendencies_0 = compute_tendencies()

        state_spect_1 = self._state_spect_tmp
        step_Euler(state_spect, dt, tendencies_0, diss, output=state_spect_1)

        phaseshift = self._get_phaseshift()
        tendencies_1_shift = compute_tendencies(phaseshift * state_spect_1)

        tendencies_d = self._state_spect_tmp
        if ts.is_transpiled:
            ts.use_block("rk2_tendencies_d")
        else:
            # transonic block (
            #     A tendencies_d, tendencies_0, tendencies_1_shift;
            #     Am1 phaseshift;
            # )
            tendencies_d[:] = 0.5 * (
                tendencies_0 + tendencies_1_shift / phaseshift
            )

        step_like_RK2(state_spect, dt, tendencies_d, diss, diss2)

    def _time_step_RK2_phaseshift_random(self):
        r"""Runge-Kutta 2 method with phase-shifting (random).

        Notes
        -----

        We consider an equation of the form

        .. math:: \p_t S = \sigma S + N(S),

        Heun's method computes an approximation of the
        solution after a time increment :math:`\dt`. We denote the
        initial time :math:`t = 0`.

        - Approximation 1:

          .. math:: \p_t \log S = \sigma + \frac{\tilde N_{0\alpha}}{S_0},

          Integrating from :math:`t` to :math:`t+\dt`, it gives:

          .. math:: S_1 = (S_0 + \tilde N_{0\alpha} \dt) e^{\sigma \dt}.

        - Approximation 2:

          .. math::
             \p_t \log S = \sigma + \frac{N_d}{S_0 e^{\sigma \frac{\dt}{2}}},

          where the dealiased non-linear term is :math:`N_d =
          (\tilde N_{0\alpha} + \tilde N_{1\beta})/2`.

          Integrating from :math:`t` to :math:`t+\dt` and retaining
          only the terms in :math:`(N\dt/S)^1` gives:

          .. math::
             S_2 = S_0 e^{\sigma \dt} + \dt N_d e^{\sigma \frac{\dt}{2}}.

        """
        dt = self.deltat
        diss, diss2 = self.exact_linear_coefs.get_updated_coefs()

        phaseshift_alpha, phaseshift_beta = self._get_phaseshift_random()

        compute_tendencies = self.sim.tendencies_nonlin
        state_spect = self.sim.state.state_spect

        tmp0 = np.empty_like(state_spect)
        state_spect_shift = mul(phaseshift_alpha, state_spect, output=tmp0)
        tendencies_0_shift = compute_tendencies(
            state_spect_shift, old=state_spect_shift
        )

        tendencies_0 = div_inplace(tendencies_0_shift, phaseshift_alpha)

        state_spect_1 = step_Euler(
            state_spect, dt, tendencies_0, diss, output=self._state_spect_tmp
        )

        tmp1 = np.empty_like(state_spect)
        state_spect_1_shift = mul(phaseshift_beta, state_spect_1, output=tmp1)
        tendencies_1_shift = compute_tendencies(
            state_spect_1_shift, old=state_spect_1_shift
        )

        tendencies_d = mean_with_phaseshift(
            tendencies_0,
            tendencies_1_shift,
            phaseshift_beta,
            output=self._state_spect_tmp,
        )
        step_like_RK2(state_spect, dt, tendencies_d, diss, diss2)

    def _time_step_RK2_phaseshift_exact(self):
        r"""Runge-Kutta 2 method with phase-shifting for exact dealiasing.

        Notes
        -----

        It requires 4 evaluations of the nonlinear terms (as RK4).

        We consider an equation of the form

        .. math:: \p_t S = \sigma S + N(S),

        Heun's method computes an approximation of the
        solution after a time increment :math:`\dt`. We denote the
        initial time :math:`t = 0`.

        - Approximation 1:

          .. math:: \p_t \log S = \sigma + \frac{N_{d0}}{S_0},

          Integrating from :math:`t` to :math:`t+\dt`, it gives:

          .. math:: S_1 = (S_0 + N_{d0} \dt) e^{\sigma \dt}.

        - Approximation 2:

          .. math::
             \p_t \log S = \sigma + \frac{N_d}{S_0 e^{\sigma \frac{\dt}{2}}},

          where the dealiased non-linear term is :math:`N_d =
          (N_{d0} +  N_{d1})/2`.

          Integrating from :math:`t` to :math:`t+\dt` and retaining
          only the terms in :math:`(N\dt/S)^1` gives:

          .. math::
             S_2 = S_0 e^{\sigma \dt} + \dt N_d e^{\sigma \frac{\dt}{2}}.

        """
        dt = self.deltat
        diss, diss2 = self.exact_linear_coefs.get_updated_coefs()
        phaseshift = self._get_phaseshift()
        compute_tendencies = self.sim.tendencies_nonlin
        state_spect = self.sim.state.state_spect

        tmp0 = np.empty_like(state_spect)
        tmp1 = np.empty_like(state_spect)
        tmp2 = np.empty_like(state_spect)
        tmp3 = np.empty_like(state_spect)

        tendencies_0 = compute_tendencies(state_spect, old=tmp0)
        state_spect_shift = mul(phaseshift, state_spect, output=tmp1)
        tendencies_0_shift = compute_tendencies(
            state_spect_shift, old=state_spect_shift
        )
        tendencies_d0 = mean_with_phaseshift(
            tendencies_0, tendencies_0_shift, phaseshift, output=tmp2
        )

        state_spect_1 = step_Euler(
            state_spect, dt, tendencies_d0, diss, output=self._state_spect_tmp
        )

        tendencies_1 = compute_tendencies(state_spect_1, old=tmp0)
        state_spect_shift = mul(phaseshift, state_spect_1, output=tmp1)
        tendencies_1_shift = compute_tendencies(
            state_spect_shift, old=state_spect_shift
        )

        tendencies_d = tmp3
        if ts.is_transpiled:
            ts.use_block("rk2_exact")
        else:
            # based on approximation 1
            # transonic block (
            #     A tendencies_d, tendencies_d0, tendencies_1, tendencies_1_shift;
            #     Am1 phaseshift
            # )

            tendencies_d[:] = 0.5 * (
                tendencies_d0
                + 0.5 * (tendencies_1 + tendencies_1_shift / phaseshift)
            )

        step_like_RK2(state_spect, dt, tendencies_d, diss, diss2)

    def _time_step_RK4(self):
        r"""Runge-Kutta 4 method.

        Notes
        -----

        .. |SA1dt2| mathmacro:: S_{A1 \mathop{dt}/2}

        We consider an equation of the form

        .. math:: \p_t S = \sigma S + N(S),

        The Runge-Kutta 4 method computes an approximation of the
        solution after a time increment :math:`\dt`. We denote the
        initial time as :math:`t = 0`. This time scheme uses 4
        approximations. Only the terms in :math:`\dt^1` are retained.

        - Approximation 1:

          .. math:: \p_t \log S = \sigma + \frac{N_0}{S_0},

          Integrating from :math:`t` to :math:`t+\dt/2` gives:

          .. math:: \SA1dt2 = (S_0 + N_0 \dt/2) e^{\sigma \frac{\dt}{2}}.

          Integrating from :math:`t` to :math:`t+\dt` gives:

          .. math:: S_{A1\dt} = (S_0 + N_0 \dt) e^{\sigma \dt}.

        - Approximation 2:

          .. math::
             \p_t \log S = \sigma
             + \frac{N(\SA1dt2)}{ \SA1dt2 },

          Integrating from :math:`t` to :math:`t+\dt/2` gives:

          .. |SA2dt2| mathmacro:: S_{A2 \mathop{dt}/2}

          .. math::
             \SA2dt2 = S_0 e^{\sigma \frac{\dt}{2}}
             + N(\SA1dt2) \frac{\dt}{2}.

          Integrating from :math:`t` to :math:`t+\dt` gives:

          .. math::
             S_{A2\dt} = S_0 e^{\sigma \dt}
             + N(\SA1dt2) e^{\sigma \frac{\dt}{2}} \dt.


        - Approximation 3:

          .. math::
             \p_t \log S = \sigma
             + \frac{N(\SA2dt2)}{ \SA2dt2 },

          Integrating from :math:`t` to :math:`t+\dt` gives:

          .. math::
             S_{A3\dt} = S_0 e^{\sigma \dt}
             + N(\SA2dt2) e^{\sigma \frac{\dt}{2}} \dt.

        - Approximation 4:

          .. math::
             \p_t \log S = \sigma
             + \frac{N(S_{A3\dt})}{ S_{A3\dt} },

          Integrating from :math:`t` to :math:`t+\dt` gives:

          .. math::
             S_{A4\dt} = S_0 e^{\sigma \dt} + N(S_{A3\dt}) \dt.

        The final result is a pondered average of the results of 4
        approximations for the time :math:`t+\dt`:

          .. math::
             \frac{1}{3} \left[
             \frac{1}{2} S_{A1\dt}
             + S_{A2\dt} + S_{A3\dt}
             + \frac{1}{2} S_{A4\dt}
             \right],

        which is equal to:

          .. math::
             S_0 e^{\sigma \dt}
             + \frac{\dt}{3} \left[
             \frac{1}{2} N_0 e^{\sigma \dt}
             + N(\SA1dt2) e^{\sigma \frac{\dt}{2}}
             + N(\SA2dt2) e^{\sigma \frac{\dt}{2}}
             + \frac{1}{2} N(S_{A3\dt})\right].

        """
        dt = self.deltat
        diss, diss2 = self.exact_linear_coefs.get_updated_coefs()

        compute_tendencies = self.sim.tendencies_nonlin
        state_spect = self.sim.state.state_spect

        tendencies_0 = compute_tendencies()
        state_spect_tmp1 = self._state_spect_tmp1

        # rk4_step0
        state_spect_tmp = step_Euler(
            state_spect, dt / 6, tendencies_0, diss, output=self._state_spect_tmp
        )
        state_spect_12_approx1 = step_Euler(
            state_spect,
            dt / 2,
            tendencies_0,
            diss2,
            output=state_spect_tmp1,
        )

        tendencies_1 = compute_tendencies(
            state_spect_12_approx1, old=tendencies_0
        )
        del state_spect_12_approx1

        state_spect_12_approx2 = state_spect_tmp1

        if ts.is_transpiled:
            ts.use_block("rk4_step1")
        else:
            # based on approximation 1
            # transonic block (
            #     A state_spect, state_spect_tmp,
            #       state_spect_12_approx2, tendencies_1;
            #     A1 diss2;
            #     float dt
            # )

            # transonic block (
            #     A state_spect, state_spect_tmp,
            #       state_spect_12_approx2, tendencies_1;
            #     A2 diss2;
            #     float dt
            # )

            state_spect_tmp[:] += dt / 3 * diss2 * tendencies_1
            state_spect_12_approx2[:] = (
                state_spect * diss2 + dt / 2 * tendencies_1
            )

        tendencies_2 = compute_tendencies(
            state_spect_12_approx2, old=tendencies_1
        )
        del state_spect_12_approx2

        state_spect_1_approx = state_spect_tmp1

        if ts.is_transpiled:
            ts.use_block("rk4_step2")
        else:
            # based on approximation 2
            # transonic block (
            #     A state_spect, state_spect_tmp,
            #       state_spect_1_approx, tendencies_2;
            #     A1 diss, diss2;
            #     float dt
            # )

            # transonic block (
            #     A state_spect, state_spect_tmp,
            #       state_spect_1_approx, tendencies_2;
            #     A2 diss, diss2;
            #     float dt
            # )

            state_spect_tmp[:] += dt / 3 * diss2 * tendencies_2
            state_spect_1_approx[:] = (
                state_spect * diss + dt * diss2 * tendencies_2
            )

        tendencies_3 = compute_tendencies(state_spect_1_approx, old=tendencies_2)
        del state_spect_1_approx

        if ts.is_transpiled:
            ts.use_block("rk4_step3")
        else:
            # result using the 4 approximations
            # transonic block (
            #     A state_spect, state_spect_tmp, tendencies_3;
            #     float dt
            # )
            state_spect[:] = state_spect_tmp + dt / 6 * tendencies_3



"""Multidimensional spectra output (:mod:`fluidsim.solvers.ns2d.output.spectra_multidim`)
=========================================================================================

.. autoclass:: SpectraMultiDimNS2D
   :members:
   :private-members:

"""

import numpy as np

from fluidsim.base.output.spectra_multidim import SpectraMultiDim


class SpectraMultiDimNS2D(SpectraMultiDim):
    """Save and plot multidimensional spectra."""

    def compute(self):
        """Compute multidimensional spectra at one time."""
        energy_fft = self.output.compute_energy_fft()

        # Computes spectrum Vs kykx
        spectrumkykx_E = self.oper.compute_spectrum_kykx(energy_fft)

        # Save results into dictionary
        dict_spectra = {"spectrumkykx_E": spectrumkykx_E}

        return dict_spectra

    def online_plot_saving(self, dict_spectra):
        raise NotImplementedError("Online plot saving is not implemented.")

    def plot(self, tmin=None, tmax=None):
        """Plots spectrumkykx averaged between tmin and tmax."""

        dict_results = self.load_mean(tmin, tmax)
        kx = dict_results["kxE"]
        ky = dict_results["kyE"]
        spectrumkykx_E = dict_results["spectrumkykx_E"]

        fig, ax = self.output.figure_axe()
        ax.set_xlabel("$k_x$")
        ax.set_ylabel("$k_y$")

        KX, KY = np.meshgrid(kx, ky)
        ax.pcolormesh(
            KX,
            KY,
            spectrumkykx_E,
            shading="nearest",
            vmin=spectrumkykx_E.min(),
            vmax=spectrumkykx_E.max(),
        )
